from abc import ABC, abstractmethod

import numpy as np
import torch

from dingo.gw.domains import Domain


class BaseFrequencyDomain(Domain, ABC):
    def __init__(self):
        self._sample_frequencies = None
        self._sample_frequencies_torch = None
        self._sample_frequencies_torch_cuda = None
        self._delta_f = None

    def __len__(self) -> int:
        """Number of frequency bins in the domain."""
        return len(self.sample_frequencies)

    def __call__(self) -> np.ndarray:
        """Return the frequency array."""
        return self.sample_frequencies

    def __getitem__(self, idx: int) -> float:
        """Index into the frequency array."""
        return self.sample_frequencies[idx]

    @property
    @abstractmethod
    def sample_frequencies(self) -> np.ndarray:
        pass

    @property
    def sample_frequencies_torch(self) -> torch.Tensor:
        if self._sample_frequencies_torch is None:
            self._sample_frequencies_torch = torch.tensor(
                self.sample_frequencies, dtype=torch.float32
            )
        return self._sample_frequencies_torch

    @property
    def sample_frequencies_torch_cuda(self) -> torch.Tensor:
        if self._sample_frequencies_torch_cuda is None:
            self._sample_frequencies_torch_cuda = self.sample_frequencies_torch.to(
                "cuda"
            )
        return self._sample_frequencies_torch_cuda

    def get_sample_frequencies_astype(
        self, data: np.ndarray | torch.Tensor
    ) -> np.ndarray | torch.Tensor:
        """
        Returns a 1D frequency array compatible with the last index of data array.

        Decides whether array is numpy or torch tensor (and cuda vs cpu).

        Parameters
        ----------
        data : Union[np.array, torch.Tensor]
            Sample data

        Returns
        -------
        Frequency array compatible with last index, of the same type as input
        """
        # Type
        if isinstance(data, np.ndarray):
            f = self.sample_frequencies
        elif isinstance(data, torch.Tensor):
            if data.is_cuda:
                f = self.sample_frequencies_torch_cuda
            else:
                f = self.sample_frequencies_torch
        else:
            raise TypeError("Invalid data type. Should be np.array or torch.Tensor.")

        return f

    @property
    @abstractmethod
    def f_min(self) -> float:
        pass

    @property
    @abstractmethod
    def frequency_mask(self) -> np.ndarray:
        pass

    @property
    @abstractmethod
    def frequency_mask_length(self) -> int:
        pass

    @property
    def delta_f(self) -> float | np.ndarray:
        return self._delta_f

    def time_translate_data(
        self, data: np.ndarray | torch.Tensor, dt: float | np.ndarray | torch.Tensor
    ) -> np.ndarray | torch.Tensor:
        r"""
        Time translate frequency-domain data by dt. Time translation corresponds (in
        frequency domain) to multiplication by
        $$
        \exp(-2 \pi i f \cdot dt).
        $$

        This method allows for multiple batch dimensions. For torch.Tensor data,
        allow for either a complex or a (real, imag) representation.

        Parameters
        ----------
        data : array-like (numpy, torch)
            Shape (B, C, N), where

                - B corresponds to any dimension >= 0,
                - C is either absent (for complex data) or has dimension >= 2 (for data
                  represented as real and imaginary parts), and
                - N is either len(self) or len(self)-self.min_idx (for truncated data).

        dt : torch tensor, or scalar (if data is numpy)
            Shape (B)

        Returns
        -------
        Array-like of the same form as data.
        """
        f = self.get_sample_frequencies_astype(data)
        if isinstance(data, np.ndarray):
            phase_shift = 2 * np.pi * np.einsum("...,i", dt, f)
        elif isinstance(data, torch.Tensor):
            # Allow for possible multiple "batch" dimensions (e.g., batch + detector,
            # which might have independent time shifts).
            phase_shift = 2 * np.pi * torch.einsum("...,i", dt, f)
        else:
            raise NotImplementedError(
                f"Time translation not implemented for data of type {data}."
            )
        return self.add_phase(data, phase_shift)

    @staticmethod
    def add_phase(
        data: np.ndarray | torch.Tensor, phase: np.ndarray | torch.Tensor
    ) -> np.ndarray | torch.Tensor:
        """
        Add a (frequency-dependent) phase to a frequency series. Allows for batching,
        as well as additional channels (such as detectors). Accounts for the fact that
        the data could be a complex frequency series or real and imaginary parts.

        Convention: the phase $\phi(f)$ is defined via $\exp[- i \phi(f)]$.

        Parameters
        ----------
        data : Union[np.array, torch.Tensor]
        phase : Union[np.array, torch.Tensor]

        Returns
        -------
        New array or tensor of the same shape as data.
        """
        if isinstance(data, np.ndarray) and np.iscomplexobj(data):
            return data * np.exp(-1j * phase)

        elif isinstance(data, torch.Tensor):
            if torch.is_complex(data):
                # Expand the trailing batch dimensions to allow for broadcasting.
                while phase.dim() < data.dim():
                    phase = phase[..., None, :]
                return data * torch.exp(-1j * phase)
            else:
                # The first two components of the second last index should be the real
                # and imaginary parts of the data. Remaining components correspond to,
                # e.g., the ASD. The "-1" below accounts for this extra dimension when
                # broadcasting.
                while phase.dim() < data.dim() - 1:
                    phase = phase[..., None, :]

                cos_phase = torch.cos(phase)
                sin_phase = torch.sin(phase)
                result = torch.empty_like(data)
                result[..., 0, :] = (
                    data[..., 0, :] * cos_phase + data[..., 1, :] * sin_phase
                )
                result[..., 1, :] = (
                    data[..., 1, :] * cos_phase - data[..., 0, :] * sin_phase
                )
                if data.shape[-2] > 2:
                    result[..., 2:, :] = data[..., 2:, :]
                return result

        else:
            raise TypeError(f"Invalid data type {type(data)}.")

    @property
    def noise_std(self) -> float:
        r"""
        Standard deviation per bin for white noise,
        $$
        \sigma_{\mathrm{noise}} = \sqrt{\frac{1}{4 \delta f}}
        $$

        Historical note: 
        Note we no longer use the noise std of the form 
        $$
        \sigma_{\mathrm{noise}} = \sqrt{\frac{w}{4 \delta f}}
        $$
        where $w$ is the window factor. For the full discussion, see 
        https://git.ligo.org/pe/pe-group-coordination/-/issues/1#note_1465275. 

        To scale noise such that it is consistent with a multivariate *unit* normal
        distribution in the, you must divide whitened data by the noise_std. For the
        UniformFrequencyDomain, noise_std is a number, as delta_f is constant across
        the domain. For the MultibandedFrequencyDomain, it is an array.
        """
        return 1 / np.sqrt(4.0 * self.delta_f)

    def check_data_compatibility(self, data: np.ndarray) -> bool:
        """
        Check that the trailing dimension of data is compatible with the domain, i.e.,
        compare against the domain length.

        Parameters
        ----------
        data : np.ndarray

        Returns
        -------
        bool
            Whether the data are compatible with domain.
        """
        if data.shape[-1] == len(self.sample_frequencies):
            return True
        else:
            return False



import os
import sys

import dingo.pipe.create_injections  # Needed for delta-function time priors.

import bilby
from bilby_pipe.input import Input
from bilby_pipe.plotting_utils import plot_whitened_data
from bilby_pipe.data_generation import DataGenerationInput as BilbyDataGenerationInput
from bilby.core.prior import PriorDict
from bilby_pipe.utils import (
    parse_args,
    logger,
    convert_string_to_dict,
    convert_prior_string_input,
    BilbyPipeError,
)
import lalsimulation as LS
import numpy as np

from dingo.core.posterior_models.build_model import build_model_from_kwargs
from dingo.gw.data.event_dataset import EventDataset
from dingo.gw.domains import UniformFrequencyDomain, build_domain_from_model_metadata
from dingo.gw.injection import Injection
from dingo.pipe.parser import create_parser

logger.name = "dingo_pipe"


class DataGenerationInput(BilbyDataGenerationInput):
    def __init__(self, args, unknown_args, create_data=True):
        # if running on the OSG, the network has been transferred to
        # the local directory, replace osdf string
        if args.osg:
            self.model = os.path.basename(args.model)
            self.model_init = os.path.basename(args.model_init)
        else:
            self.model = args.model
            self.model_init = args.model_init

        Input.__init__(self, args, unknown_args)
        # Generic initialisation
        self.meta_data = dict(
            command_line_args=args.__dict__,
            unknown_command_line_args=unknown_args,
            injection_parameters=None,
            # bilby_version=bilby.__version__,
            # bilby_pipe_version=get_version_information(),
        )
        self.injection_parameters = None

        # Admin arguments
        self.ini = args.ini
        self.transfer_files = args.transfer_files

        # Global settings
        # self.cosmology = args.cosmology

        # Run index arguments
        self.idx = args.idx
        self.generation_seed = args.generation_seed
        self.trigger_time = args.trigger_time

        # Naming arguments
        self.outdir = args.outdir
        self.label = args.label

        # Prior arguments
        # self.reference_frame = args.reference_frame
        # self.time_reference = "geocent" # DINGO mod used for saving data dump
        # self.phase_marginalization = args.phase_marginalization
        self.prior_dict = args.prior_dict
        self.default_prior = "BBHPriorDict"
        self.time_reference = "geocent"
        self.prior_dict_updates = args.prior_dict_updates

        # Whether to generate data for importance sampling. This must be done when
        # desired data settings differ from those used for network training. If this is
        # the case, save the new data to a different file name.

        self.importance_sampling = args.importance_sampling_generation
        self.importance_sampling_updates = args.importance_sampling_updates
        if self.importance_sampling:
            # Updates to frequency range should not affect the data generation for importance sampling
            if "minimum_frequency" in self.importance_sampling_updates:
                self.importance_sampling_updates.pop("minimum_frequency")
            if "maximum_frequency" in self.importance_sampling_updates:
                self.importance_sampling_updates.pop("maximum_frequency")
            vars(args).update(self.importance_sampling_updates)

        # Data arguments
        self.ignore_gwpy_data_quality_check = args.ignore_gwpy_data_quality_check
        self.detectors = args.detectors
        self.channel_dict = args.channel_dict
        self.fetch_open_data_kwargs = args.fetch_open_data_kwargs
        self.data_dict = args.data_dict
        self.data_format = args.data_format
        self.allow_tape = args.allow_tape
        self.tukey_roll_off = args.tukey_roll_off
        self.gaussian_noise = args.gaussian_noise
        self.zero_noise = args.zero_noise
        self.resampling_method = args.resampling_method

        if args.timeslide_dict is not None:
            self.timeslide_dict = convert_string_to_dict(args.timeslide_dict)
            logger.info(f"Read-in timeslide dict directly: {self.timeslide_dict}")
        elif args.timeslide_file is not None:
            self.gps_file = args.gps_file
            self.timeslide_file = args.timeslide_file
            self.timeslide_dict = self.get_timeslide_dict(self.idx)

        # Data duration arguments
        self.duration = args.duration
        self.post_trigger_duration = args.post_trigger_duration

        # Frequencies
        self.sampling_frequency = args.sampling_frequency
        self.minimum_frequency = args.minimum_frequency
        self.maximum_frequency = args.maximum_frequency
        self.reference_frequency = args.reference_frequency

        # Waveform, source model and likelihood
        self.waveform_generator_class = (
            "bilby.gw.waveform_generator.LALCBCWaveformGenerator"
        )
        # self.waveform_generator_class_ctor_args = (
        #     args.waveform_generator_constructor_dict
        # )
        self.waveform_approximant = args.waveform_approximant
        self.catch_waveform_errors = args.catch_waveform_errors
        # TODO: These are set to parser defaults. Fix to set from model.
        self.pn_spin_order = -1
        self.pn_tidal_order = -1
        self.pn_phase_order = -1
        self.pn_amplitude_order = 0
        self.mode_array = None
        # don't set self.waveform_arguments_dict, it will be updated later by injection_waveform_arguments
        self.waveform_arguments_dict = None
        self.injection_waveform_arguments = args.injection_waveform_arguments
        self.numerical_relativity_file = args.numerical_relativity_file
        self.dingo_injection = args.dingo_injection
        self.injection_waveform_approximant = args.injection_waveform_approximant
        if args.injection_waveform_approximant in [
            "SEOBNRv5PHM",
            "SEOBNRv5EHM",
            "SEOBNRv5HM",
        ]:
            self.injection_frequency_domain_source_model = "gwsignal_binary_black_hole"
            self.frequency_domain_source_model = "gwsignal_binary_black_hole"
        else:
            self.injection_frequency_domain_source_model = "lal_binary_black_hole"
            self.frequency_domain_source_model = "lal_binary_black_hole"

        # DINGO mod
        self.save_bilby_data_dump = args.save_bilby_data_dump
        if self.save_bilby_data_dump:
            self.time_reference = args.time_reference
            self.deltaT = args.deltaT

        # self.conversion_function = args.conversion_function
        # self.generation_function = args.generation_function
        # self.likelihood_type = args.likelihood_type
        # self.extra_likelihood_kwargs = args.extra_likelihood_kwargs
        self.enforce_signal_duration = args.enforce_signal_duration

        # PSD
        self.psd_maximum_duration = args.psd_maximum_duration
        self.psd_dict = args.psd_dict
        self.psd_length = args.psd_length
        self.psd_fractional_overlap = args.psd_fractional_overlap
        self.psd_start_time = args.psd_start_time
        self.psd_method = args.psd_method

        # # ROQ
        # self.roq_folder = args.roq_folder
        # self.roq_linear_matrix = args.roq_linear_matrix
        # self.roq_quadratic_matrix = args.roq_quadratic_matrix
        # self.roq_weights = args.roq_weights
        # self.roq_weight_format = args.roq_weight_format
        # self.roq_scale_factor = args.roq_scale_factor
        #
        # # Calibration
        # self.calibration_model = args.calibration_model
        # self.calibration_correction_type = args.calibration_correction_type
        # self.spline_calibration_envelope_dict = args.spline_calibration_envelope_dict
        # self.spline_calibration_amplitude_uncertainty_dict = (
        #     args.spline_calibration_amplitude_uncertainty_dict
        # )
        # self.spline_calibration_phase_uncertainty_dict = (
        #     args.spline_calibration_phase_uncertainty_dict
        # )
        # self.spline_calibration_nodes = args.spline_calibration_nodes
        # self.calibration_prior_boundary = args.calibration_prior_boundary
        #
        # # Marginalization
        # self.distance_marginalization = args.distance_marginalization
        # self.distance_marginalization_lookup_table = (
        #     args.distance_marginalization_lookup_table
        # )
        # self.phase_marginalization = args.phase_marginalization
        # self.time_marginalization = args.time_marginalization
        # self.jitter_time = args.jitter_time
        #
        # # Plotting
        self.plot_data = args.plot_data
        self.plot_spectrogram = args.plot_spectrogram
        self.plot_injection = args.plot_injection

        if create_data:
            if self.dingo_injection:
                self.create_data_dingo_injection(args)
            else:
                self.create_data(args)

    def create_data_dingo_injection(self, args):
        """Adaptation of create_data to use Dingo signal models rather than Bilby.

        First, executes create_data but without any requested injections. This creates
        a  noise-only dataset.

        Second, calls _inject_dingo_signal to generate the Dingo signal waveform and
        add it to the noisy data within the interferometers.
        """
        # Save values of relevant args.
        injection = args.injection
        injection_file = args.injection_file
        injection_dict = args.injection_dict

        args.injection = False
        args.injection_file = None
        args.injection_dict = None

        # Create noise.
        self.create_data(args)

        # Reset args.
        args.injection = injection
        args.injection_file = injection_file
        args.injection_dict = injection_dict
        self.injection = args.injection
        self.injection_file = args.injection_file
        self.injection_dict = args.injection_dict

        if self.injection:
            self._inject_dingo_signal(args)

    @BilbyDataGenerationInput.interferometers.setter
    def interferometers(self, interferometers):
        # Monkey patch to avoid bilby_pipe zeroing data below
        # self.minimum_frequency_dict and above self.maximum_frequency_dict. Remove
        # this code if we are happy to let bilby_pipe do this. Possible issue is edge
        # effects for multi-banded frequency domain, so instead we do the masking
        # within Dingo.
        for ifo in interferometers:
            if isinstance(ifo, bilby.gw.detector.Interferometer) is False:
                raise BilbyPipeError(f"ifo={ifo} is not a bilby Interferometer")
            # if self.minimum_frequency is not None:
            #     ifo.minimum_frequency = self.minimum_frequency_dict[ifo.name]
            # if self.maximum_frequency is not None:
            #     ifo.maximum_frequency = self.maximum_frequency_dict[ifo.name]
            # if self.calibration_model is not None:
            #     self.add_calibration_model_to_interferometers(ifo)

        self._interferometers = interferometers
        self.data_set = True
        if self.plot_data:
            interferometers.plot_data(outdir=self.data_directory, label=self.label)
            plot_whitened_data(
                interferometers=interferometers,
                data_directory=self.data_directory,
                label=self.label,
            )

    def _inject_dingo_signal(self, args):
        """Generate a GW signal using the dingo.gw.injection class and add it to the
        interferometer strain data. Also compute SNRs and store them."""
        try:
            model = build_model_from_kwargs(
                filename=args.model, device="meta", load_training_info=False
            )
        except RuntimeError:
            # 'meta' is not supported by older version of python / torch
            model = build_model_from_kwargs(
                filename=args.model, device="cpu", load_training_info=False
            )

        injection = Injection.from_posterior_model_metadata(model.metadata)
        injection.use_base_domain = True  # Do not generate MFD signals.
        injection.t_ref = self.trigger_time
        injection._initialize_transform()

        # Possibly update waveform generator based on supplied settings. Note that
        # default values will have been set by dingo_pipe from the DINGO model.
        waveform_arguments = self.get_injection_waveform_arguments()
        injection.f_ref = waveform_arguments["reference_frequency"]
        injection.waveform_generator.approximant = LS.GetApproximantFromString(
            waveform_arguments["waveform_approximant"]
        )
        injection.waveform_generator.approximant_str = waveform_arguments[
            "waveform_approximant"
        ]

        logger.info("Injecting waveform from DINGO with ")
        logger.info(f"data_domain = {injection.data_domain.domain_dict}")
        for prop in [
            "t_ref",
        ]:
            logger.info(f"{prop} = {getattr(injection, prop)}")
        for prop in [
            "approximant_str",
            "f_ref",
            "f_start",
        ]:
            logger.info(f"{prop} = {getattr(injection.waveform_generator, prop)}")

        # Generate signal
        self.injection_parameters = self.injection_df.iloc[self.idx].to_dict()
        theta = self.injection_parameters.copy()
        theta["geocent_time"] -= self.trigger_time
        signal = injection.signal(theta)

        # Add signal to interferometer data
        domain = injection.data_domain
        for ifo in self.interferometers:
            s = signal["waveform"][ifo.name]
            s = domain.time_translate_data(s, -self.post_trigger_duration)
            # Interferometer data extends to Nyquist = f_max / 2, so pad signal. Will
            # be truncated when saved as HDF5.
            s = np.pad(
                s, (0, len(ifo.strain_data.frequency_domain_strain) - len(domain))
            )
            ifo.strain_data.frequency_domain_strain += s
            ifo.meta_data["optimal_SNR"] = np.sqrt(
                ifo.optimal_snr_squared(signal=s).real
            )
            ifo.meta_data["matched_filter_SNR"] = ifo.matched_filter_snr(signal=s)

    def save_hdf5(self):
        """
        Save frequency-domain strain and ASDs as DingoDataset HDF5 format.

        This method will also save the PSDs as .txt files in the data directory
        for easy reading by pesummary and Bilby.
        """

        try:
            model = build_model_from_kwargs(
                filename=self.model, device="meta", load_training_info=False
            )
        except RuntimeError:
            # 'meta' is not supported by older version of python / torch
            model = build_model_from_kwargs(
                filename=self.model, device="cpu", load_training_info=False
            )
        domain = build_domain_from_model_metadata(model.metadata, base=True)
        assert isinstance(domain, UniformFrequencyDomain)

        if self.save_bilby_data_dump:
            # this is needed because we want bilby to use the updated DINGO
            # prior
            self.waveform_arguments_dict = self.injection_waveform_arguments
            self.prior_dict = self.priors
            self.likelihood_type = "GravitationalWaveTransient"
            self.calibration_marginalization = False
            self.phase_marginalization = False
            self.time_marginalization = False
            self.distance_marginalization = False
            self.number_of_response_curves = 0
            self._distance_marginalization_lookup_table = None
            self.reference_frame = "sky"
            self.fiducial_parameters = None
            self.update_fiducial_parameters = None
            self.epsilon = None
            self.jitter_time = True
            self.save_data_dump()

        # PSD and strain data.
        data = {"waveform": {}, "asds": {}}  # TODO: Rename these keys.
        for ifo in self.interferometers:
            strain = ifo.strain_data.frequency_domain_strain
            frequency_array = ifo.strain_data.frequency_array
            asd = ifo.power_spectral_density.get_amplitude_spectral_density_array(
                frequency_array
            )

            # These arrays extend up to self.sampling_frequency / 2. Truncate them to
            # the model maximum frequency, and also set the ASD to 1.0 below model
            # minimum frequency.
            strain = domain.update_data(strain)
            asd = domain.update_data(asd, low_value=1.0)

            # Dingo expects data to have trigger time 0, so we apply a cyclic time shift
            # by the post-trigger duration.
            strain = domain.time_translate_data(strain, self.post_trigger_duration)

            # Note that the ASD estimated by the Bilby Interferometer differs ever so
            # slightly from the ASDs we computed before using pycbc. In addition,
            # there is no handling of NaNs in the strain data from which the ASD is
            # estimated.

            data["waveform"][ifo.name] = strain
            data["asds"][ifo.name] = asd

        # Data conditioning settings.
        settings = {
            "time_event": self.trigger_time,
            "time_buffer": self.post_trigger_duration,
            "detectors": self.detectors,
            "f_s": self.sampling_frequency,
            "T": self.duration,
            "window_type": "tukey",
            "roll_off": self.tukey_roll_off,
            "minimum_frequency": self.minimum_frequency_dict,
            "maximum_frequency": self.maximum_frequency_dict,
        }

        for k in [
            "psd_duration",
            "psd_dict",
            "psd_fractional_overlap",
            "psd_start_time",
            "psd_method",
            "channel_dict",
            "data_dict",
            "injection",
            "generation_seed",
            "gaussian_noise",
            "zero_noise",
            "numerical_relativity_file",
            "injection_waveform_approximant",
            "injection_frequency_domain_source_model",
        ]:
            try:
                v = getattr(self, k)
            except AttributeError:
                continue
            if v is not None and v is not False:
                settings[k] = v

        if self.injection:
            settings["injection_parameters"] = self.injection_parameters.copy()
            settings["dingo_injection"] = self.dingo_injection
            # Dingo and Bilby have different geocent_time conventions.
            settings["injection_parameters"]["geocent_time"] -= self.trigger_time
            settings["optimal_SNR"] = {
                k: v["optimal_SNR"].item()
                if hasattr(v["optimal_SNR"], "item")
                else v["optimal_SNR"]
                for k, v in self.interferometers.meta_data.items()
            }
            settings["matched_filter_SNR"] = {
                k: v["matched_filter_SNR"].item()
                if hasattr(v["matched_filter_SNR"], "item")
                else v["matched_filter_SNR"]
                for k, v in self.interferometers.meta_data.items()
            }

        dataset = EventDataset(
            dictionary={
                "data": data,
                "settings": settings,
            }
        )
        dataset.to_file(self.event_data_file)

        # also saving the psd as a .txt file which can be read in
        # easily by pesummary or bilby
        for ifo in self.interferometers:
            np.savetxt(
                os.path.join(self.data_directory, f"{ifo.name}_psd.txt"),
                np.vstack([domain(), data["asds"][ifo.name] ** 2]).T,
            )

    @property
    def event_data_file(self):
        return os.path.join(
            self.data_directory, "_".join([self.label, "event_data.hdf5"])
        )

    @property
    def importance_sampling_updates(self):
        return self._importance_sampling_updates

    @importance_sampling_updates.setter
    def importance_sampling_updates(self, setting):
        if setting is not None:
            self._importance_sampling_updates = convert_string_to_dict(
                setting, "importance-sampling-updates"
            )
        else:
            self._importance_sampling_updates = None

    @property
    def prior_dict_updates(self):
        """The input prior_dict from the ini (if given)

        Note, this is not the bilby prior (see self.priors for that), this is
        a key-val dictionary where the val's are strings which are converting
        into bilby priors in `_get_prior
        """
        return self._prior_dict_updates

    @prior_dict_updates.setter
    def prior_dict_updates(self, prior_dict_updates):
        if isinstance(prior_dict_updates, dict):
            prior_dict_updates = prior_dict_updates
        elif isinstance(prior_dict_updates, str):
            prior_dict_updates = convert_prior_string_input(prior_dict_updates)
        elif prior_dict_updates is None:
            self._prior_dict_updates = None
            return
        else:
            raise BilbyPipeError(
                f"prior_dict_updates={prior_dict_updates} not " f"understood"
            )

        self._prior_dict_updates = {
            self._convert_prior_dict_key(key): val
            for key, val in prior_dict_updates.items()
        }

    def _get_priors(self, add_time=True):
        priors = super()._get_priors(add_time=add_time)
        priors.update(PriorDict(self.prior_dict_updates))
        return priors


def create_generation_parser():
    """Data generation parser creation"""
    return create_parser(top_level=False)


def main():
    """Data generation main logic"""
    args, unknown_args = parse_args(sys.argv[1:], create_generation_parser())
    # log_version_information()
    data = DataGenerationInput(args, unknown_args)
    data.save_hdf5()
    logger.info("Completed data generation")



from .base_model import BasePosteriorModel

from dingo.core.nn.nsf import (
    create_nsf_with_rb_projection_embedding_net,
    create_nsf_wrapped,
)


class NormalizingFlowPosteriorModel(BasePosteriorModel):
    """
    Posterior model based on a (discrete) normalizing flow.

    A normalizing flow describes a distribution as a sequence of discrete
    transformations on a parameter space, ultimately taking samples from the base space
    (multivariate standard normal) to the desired distribution. The discrete transforms
    are parametrized functions (e.g., splines), which are designed to be invertible with
    simple Jacobian determinant. The probability density is given by the change of
    variables rule,

    q(theta | d) = pi(f_d^{-1}(theta)) | det J_{f_d^{-1}} |

    where
        pi = N(0,1)^D is the base space distribution
        f_d is the normalizing flow on the D-dimensional space

    The flow f_d is allowed to depend on context information d, which would be
    observational data in the case of posterior estimation. By construction, the flow
    has fast sampling and density evaluation, require just forward passes of the network.

    This class uses normalizing flows from the dingo.core.nn.nsf module (which in turn
    uses glasflow, which is based on nflows). It is intended to construct and hold a
    neural network for estimating the posterior density, as well as saving / loading,
    and training. It also calls the sampling and density evaluation routines from the
    flows.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def initialize_network(self):
        model_kwargs = {
            k: v for k, v in self.model_kwargs.items() if k != "posterior_model_type"
        }
        if self.initial_weights is not None:
            model_kwargs["initial_weights"] = self.initial_weights

        if self.model_kwargs.get("embedding_kwargs", False):
            self.network = create_nsf_with_rb_projection_embedding_net(**model_kwargs)
        else:
            self.network = create_nsf_wrapped(**model_kwargs["posterior_kwargs"])

    def log_prob(self, theta, *context):
        return self.network(theta, *context)

    def sample(self, *context, num_samples: int = 1):
        return self.network.sample(*context, num_samples=num_samples)

    def sample_and_log_prob(self, *context, num_samples: int = 1):
        return self.network.sample_and_log_prob(*context, num_samples=num_samples)

    def loss(self, theta, *context):
        return -self.network(theta, *context).mean()



from typing import Iterable, Union, Optional
import numpy as np
import torch
from copy import copy

from .base_frequency_domain import BaseFrequencyDomain
from .uniform_frequency_domain import UniformFrequencyDomain


class MultibandedFrequencyDomain(BaseFrequencyDomain):
    r"""
    Defines a non-uniform frequency domain that is made up of a sequence of
    uniform-frequency domain bands. Each subsequent band in the sequence has double the
    bin-width of the previous one, i.e., delta_f is doubled each band as one moves up
    the bands. This is intended to allow for efficient representation of gravitational
    waveforms, which generally have slower oscillations at higher frequencies. Indeed,
    the leading order chirp has phase evolution [see
    https://doi.org/10.1103/PhysRevD.49.2658],
    $$
    \Psi(f) = \frac{3}{4}(8 \pi \mathcal{M} f)^{-5/3},
    $$
    hence a coarser grid can be used at higher f.

    The domain is partitioned into bands via a sequence of nodes that are specified at
    initialization.

    In comparison to the UniformFrequencyDomain, the MultibandedFrequencyDomain has the
    following key differences:

    * The sample frequencies start at the first node, rather than f = 0.0 Hz.

    * Quantities such as delta_f, noise_std, etc., are represented as arrays rather than
    scalars, as they vary depending on f.

    The MultibandedFrequencyDomain furthermore has an attribute base_domain,
    which holds an underlying UniformFrequencyDomain object. The decimate() method
    decimates data in the base_domain to the multi-banded domain.
    """

    def __init__(
        self,
        nodes: Iterable[float],
        delta_f_initial: float,
        base_domain: Union[UniformFrequencyDomain, dict],
    ):
        """
        Parameters
        ----------
        nodes: Iterable[float]
            Defines the partitioning of the underlying frequency domain into bands. In
            total, there are len(nodes) - 1 frequency bands. Band j consists of
            decimated data from the base domain in the range [nodes[j]:nodes[j+1]).
        delta_f_initial: float
            delta_f of band 0. The decimation factor doubles between adjacent bands,
            so delta_f is doubled as well.
        base_domain: Union[UniformFrequencyDomain, dict]
            Original (uniform frequency) domain of data, which is the starting point
            for the decimation. This determines the decimation details and the noise_std.
            Either provided as dict for build_domain, or as domain_object.
        """
        super().__init__()
        if isinstance(base_domain, dict):
            from .build_domain import build_domain

            base_domain = build_domain(base_domain)

        self.nodes = np.array(nodes, dtype=np.float32)
        self.base_domain = base_domain
        self._initialize_bands(delta_f_initial)
        if not isinstance(self.base_domain, UniformFrequencyDomain):
            raise ValueError(
                f"Expected domain type UniformFrequencyDomain, got {type(base_domain)}."
            )
        # truncation indices for domain update
        self._range_update_idx_lower = None
        self._range_update_idx_upper = None
        self._range_update_initial_length = None

    def _initialize_bands(self, delta_f_initial: float):
        if len(self.nodes.shape) != 1:
            raise ValueError(
                f"Expected format [num_bands + 1] for nodes, "
                f"got {self.nodes.shape}."
            )
        self.num_bands = len(self.nodes) - 1
        self._nodes_indices = (self.nodes / self.base_domain.delta_f).astype(int)

        self._delta_f_bands = (
            delta_f_initial * (2 ** np.arange(self.num_bands))
        ).astype(np.float32)
        self._decimation_factors_bands = (
            self._delta_f_bands / self.base_domain.delta_f
        ).astype(int)
        self._num_bins_bands = (
            (self._nodes_indices[1:] - self._nodes_indices[:-1])
            / self._decimation_factors_bands
        ).astype(int)

        self._band_assignment = np.concatenate(
            [
                np.ones(num_bins_band, dtype=int) * idx
                for idx, num_bins_band in enumerate(self._num_bins_bands)
            ]
        )
        self._delta_f = self._delta_f_bands[self._band_assignment]

        # For each bin, [self._f_base_lower, self._f_base_upper] describes the
        # frequency range in the base domain which is used for truncation.
        self._f_base_lower = np.concatenate(
            (self.nodes[:1], self.nodes[0] + np.cumsum(self._delta_f[:-1]))
        )
        self._f_base_upper = (
            self.nodes[0] + np.cumsum(self._delta_f) - self.base_domain.delta_f
        )

        # Set sample frequencies as mean of decimation range.
        self._sample_frequencies = (self._f_base_upper + self._f_base_lower) / 2
        self._sample_frequencies_torch = None
        self._sample_frequencies_torch_cuda = None
        # sample_frequencies should always be the decimation of the base domain
        # frequencies.

        if self.f_min not in self.base_domain() or self.f_max not in self.base_domain():
            raise ValueError(
                f"Endpoints ({self.f_min}, {self.f_max}) not in base "
                f"domain, {self.base_domain.domain_dict}"
            )

        # Note that f_max from the base domain can differ from that of the multi-banded
        # domain. This occurs as a boundary effect, since a fixed number of bins in the
        # base domain are averaged to obtain a bin in the multi-banded domain. When
        # decimating, any extra bins in the base domain are dropped.

    def decimate(self, data: np.ndarray | torch.Tensor) -> np.ndarray | torch.Tensor:
        """
        Decimate data from the base_domain to the multi-banded domain.

        Parameters
        ----------
        data : array-like (np.ndarray or torch.Tensor)
            Decimation is done along the trailing dimension of this array. This
            dimension should therefore be compatible with the base frequency domain,
            i.e., running from 0.0 Hz or f_min up to f_max, with uniform delta_f.

        Returns
        -------
        Decimated array of the same type as the input.
        """
        if data.shape[-1] == len(self.base_domain):
            offset_idx = 0
        elif data.shape[-1] == len(self.base_domain) - self.base_domain.min_idx:
            offset_idx = -self.base_domain.min_idx
        else:
            raise ValueError(
                f"Provided data has {data.shape[-1]} bins, which is incompatible with "
                f"the expected domain of length {len(self.base_domain)}"
            )
        if isinstance(data, np.ndarray):
            data_decimated = np.empty((*data.shape[:-1], len(self)), dtype=data.dtype)
        elif isinstance(data, torch.Tensor):
            data_decimated = torch.empty(
                (*data.shape[:-1], len(self)), dtype=data.dtype
            )
        else:
            raise NotImplementedError(
                f"Decimation not implemented for data of type {data}."
            )

        lower_out = 0  # running index for decimated band data
        for idx_band in range(self.num_bands):
            lower_in = self._nodes_indices[idx_band] + offset_idx
            upper_in = self._nodes_indices[idx_band + 1] + offset_idx
            decimation_factor = self._decimation_factors_bands[idx_band]
            num_bins = self._num_bins_bands[idx_band]

            data_decimated[..., lower_out : lower_out + num_bins] = decimate_uniform(
                data[..., lower_in:upper_in], decimation_factor
            )
            lower_out += num_bins

        assert lower_out == len(self)

        return data_decimated

    def update(self, new_settings: dict):
        """
        Update the domain by truncating the frequency range (by specifying new f_min,
        f_max).

        After calling this function, data from the original domain can be truncated to
        the new domain using self.update_data(). For simplicity, we do not allow for
        multiple updates of the domain.

        Parameters
        ----------
        new_settings : dict
            Settings dictionary. Keys must either be the keys contained in domain_dict, or
            a subset of ["f_min", "f_max"].
        """
        if set(new_settings.keys()).issubset(["f_min", "f_max"]):
            self._set_new_range(**new_settings)
        elif set(new_settings.keys()) == self.domain_dict.keys():
            if new_settings == self.domain_dict:
                return
            self._set_new_range(
                f_min=new_settings["base_domain"]["f_min"],
                f_max=new_settings["base_domain"]["f_max"],
            )
            if self.domain_dict != new_settings:
                raise ValueError(
                    f"Update settings {new_settings} are incompatible with "
                    f"domain settings {self.domain_dict}."
                )
        else:
            raise ValueError(
                f"Invalid argument for domain update {new_settings}. Must either be "
                f'{list(self.domain_dict.keys())} or a subset of ["f_min, f_max"].'
            )

    def _set_new_range(
        self, f_min: Optional[float] = None, f_max: Optional[float] = None
    ):
        """
        Set a new range [f_min, f_max] for the domain. This operation is only allowed
        if the new range is contained within the old one.

        Note: f_min, f_max correspond to the range in the *base_domain*.

        Parameters
        ----------
        f_min : float
            New minimum frequency (optional).
        f_max : float
            New maximum frequency (optional).
        """
        if f_min is None and f_max is None:
            return
        if self._range_update_initial_length is not None:
            raise ValueError(f"Can't update domain of type {type(self)} a second time.")
        if f_min is not None and f_max is not None and f_min >= f_max:
            raise ValueError("f_min must not be larger than f_max.")

        lower_bin, upper_bin = 0, len(self) - 1

        if f_min is not None:
            if self._f_base_lower[0] <= f_min <= self._f_base_lower[-1]:
                # find new starting bin (first element with f >= f_min)
                lower_bin = np.where(self._f_base_lower >= f_min)[0][0]
            else:
                raise ValueError(
                    f"f_min = {f_min} is not in expected range "
                    f"[{self._f_base_lower[-1], self._f_base_lower[-1]}]."
                )

        if f_max is not None:
            if self._f_base_upper[0] <= f_max <= self._f_base_upper[-1]:
                # find new final bin (last element where f <= f_max)
                upper_bin = np.where(self._f_base_upper <= f_max)[0][-1]
            else:
                raise ValueError(
                    f"f_min = {f_min} is not in expected range "
                    f"[{self._f_base_lower[-1], self._f_base_lower[-1]}]."
                )

        lower_band = self._band_assignment[lower_bin]
        upper_band = self._band_assignment[upper_bin]
        # new nodes extend to upper_band + 2: we have +1 from the exclusive end index
        # and +1, as we have num_bands + 1 elements in nodes
        nodes_new = copy(self.nodes)[lower_band : upper_band + 2]
        nodes_new[0] = self._f_base_lower[lower_bin]
        nodes_new[-1] = self._f_base_upper[upper_bin] + self.base_domain.delta_f

        # Update base_domain f_min and f_max. These values might differ slightly from
        # the final values for the multi-banded domain due to edge effects. Note that
        # we do this update *after* all of our validation checks but *before* changing
        # the state of the class.
        self.base_domain.update({"f_min": f_min, "f_max": f_max})

        self._range_update_initial_length = len(self)
        self._range_update_idx_lower = lower_bin
        self._range_update_idx_upper = upper_bin

        self.nodes = nodes_new
        self._initialize_bands(self._delta_f_bands[lower_band])
        assert self._range_update_idx_upper - self._range_update_idx_lower + 1 == len(
            self
        )

        assert self.base_domain.f_min <= self.f_min
        assert self.base_domain.f_max >= self.f_max

    def update_data(
        self, data: np.ndarray | torch.Tensor, axis: int = -1, **kwargs
    ) -> np.ndarray | torch.Tensor:
        """
        Truncates the data array to be compatible with the domain. This is used when
        changing f_min or f_max.

        update_data() will only have an effect after updating the domain to have a new
        frequency range using self.update().

        Parameters
        ----------
        data : array-like (np.ndarray or torch.Tensor)
            Array should be compatible with either the original or updated
            MultibandedFrequencyDomain along the specified axis. In the latter
            case, nothing is done. In the former, data are truncated appropriately.
        axis: int
            Axis along which to operate.

        Returns
        -------
        Updated data of the same type as input.
        """
        if data.shape[axis] == len(self):
            return data
        elif (
            self._range_update_initial_length is not None
            and data.shape[axis] == self._range_update_initial_length
        ):
            sl = [slice(None)] * data.ndim
            # First truncate beyond f_max.
            sl[axis] = slice(
                self._range_update_idx_lower, self._range_update_idx_upper + 1
            )
            data = data[tuple(sl)]
            return data
        else:
            raise ValueError(
                f"Data (shape {data.shape}) incompatible with the domain "
                f"(length {len(self)})."
            )

    @property
    def sample_frequencies(self) -> np.ndarray:
        return self._sample_frequencies

    @property
    def frequency_mask(self) -> np.ndarray:
        """Array of len(self) consisting of ones.

        As the MultibandedFrequencyDomain starts from f_min, no masking is generally
        required."""
        return np.ones_like(self.sample_frequencies)

    @property
    def frequency_mask_length(self) -> int:
        return len(self.frequency_mask)

    @property
    def min_idx(self):
        return 0

    @property
    def max_idx(self):
        return len(self) - 1

    @property
    def f_max(self) -> float:
        return float(self._f_base_upper[-1])

    @property
    def f_min(self) -> float:
        return float(self._f_base_lower[0])

    @property
    def duration(self) -> float:
        raise NotImplementedError()

    @property
    def sampling_rate(self) -> float:
        raise NotImplementedError()

    @property
    def domain_dict(self) -> dict:
        """Enables to rebuild the domain via calling build_domain(domain_dict)."""
        # Call tolist() on self.bands, such that it can be saved as str for metadata.
        return {
            "type": "MultibandedFrequencyDomain",
            "nodes": self.nodes.tolist(),
            "delta_f_initial": self._delta_f_bands[0].item(),
            "base_domain": self.base_domain.domain_dict,
        }


######################
### util functions ###
######################


def decimate_uniform(data, decimation_factor: int):
    """
    Reduce dimension of data by decimation_factor along last axis, by uniformly
    averaging sets of decimation_factor neighbouring bins.

    Parameters
    ----------
    data
        Array or tensor to be decimated.
    decimation_factor
        Factor by how much to compress. Needs to divide data.shape[-1].
    Returns
    -------
    data_decimated
        Uniformly decimated data, as array or tensor.
        Shape (*data.shape[:-1], data.shape[-1]/decimation_factor).
    """
    if data.shape[-1] % decimation_factor != 0:
        raise ValueError(
            f"data.shape[-1] ({data.shape[-1]} is not a multiple of decimation_factor "
            f"({decimation_factor})."
        )
    if isinstance(data, np.ndarray):
        return (
            np.sum(np.reshape(data, (*data.shape[:-1], -1, decimation_factor)), axis=-1)
            / decimation_factor
        )
    elif isinstance(data, torch.Tensor):
        return (
            torch.sum(
                torch.reshape(data, (*data.shape[:-1], -1, decimation_factor)), dim=-1
            )
            / decimation_factor
        )
    else:
        raise NotImplementedError(
            f"Decimation not implemented for data of type {data}."
        )



import inspect
import textwrap

import torch

from .cflow_base import ContinuousFlowPosteriorModel


class ScoreDiffusionPosteriorModel(ContinuousFlowPosteriorModel):
    __doc__ = (
        inspect.getdoc(ContinuousFlowPosteriorModel)
        + "\n\n"
        + textwrap.dedent(
            """\
                Training with score matching:
            
                    t               ~ U[0, 1-eps)                               noise level
                    theta_0         ~ N(0, 1)                                   sampled noise
                    theta_1         = theta                                     pure sample
                    theta_t         = c1(t) * theta_1 + c0(t) * theta_0         noisy sample
            
                    eps             > 0
                    c0              = sigma(t)
                    c1              = alpha(1-t)
            
                    score_target    = theta_0 / sigma_t
                    weight          = 1/2 * {score-matching: sigma(t)^2, score-flow: beta(1-t), ...}
                    loss            = || score_target - network(theta_t, t) ||
                    
                To specify the score matching model, "posterior_kwargs" should 
                additionally specify the noise properties used for the diffusion (
                beta_min, beta_max, epsilon).
                """
        )
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.eps = self.model_kwargs["posterior_kwargs"]["epsilon"]
        self.beta_min = self.model_kwargs["posterior_kwargs"]["beta_min"]
        self.beta_max = self.model_kwargs["posterior_kwargs"]["beta_max"]

        likelihood_weighting = self.model_kwargs["posterior_kwargs"].get(
            "likelihood_weighting", "score-matching"
        )
        if likelihood_weighting:
            self.likelihood_weighting = self.get_likelihood_weighting(
                likelihood_weighting
            )

    def loss(self, theta, *context_data):
        """
        Returns the score matching loss for parameters theta conditioned on context.

        Parameters
        ----------
        theta: torch.tensor
            parameters (e.g., binary-black hole parameters)
        *context_data: list[torch.Tensor]
            context data (e.g., gravitational-wave data)

        Returns
        -------
        torch.tensor
            Loss.
        """
        t, theta_t, score = self.get_t_theta_t_score(theta_1=theta)
        pred_score = self.network(t, theta_t, *context_data)

        weighting = self.likelihood_weighting(t)
        losses = torch.square(pred_score - score)
        losses = 1 / 2 * torch.sum(losses, dim=1) * weighting
        loss = torch.mean(losses)
        return loss

    def evaluate_vector_field(self, t, theta_t, *context_data):
        """
        Evaluate the vector field v(t, theta_t, context_data) that generates the flow
        via the ODE

            d/dt f(theta_t, t, context) = v(f(theta_t, t, context), t, context).

        For score matching, the vector field (or drift function) is computed
        from the predicted score.

        Parameters
        ----------
        t: float
            time (noise level)
        theta_t: torch.Tensor
            noisy parameters, perturbed with noise level t
        *context_data: list[torch.tensor]
            list with context data (GW data)
        """
        # If t is a number (and thus the same for each element in this batch),
        # expand it as a tensor. This is required for the odeint solver.
        t = t * torch.ones(len(theta_t), device=theta_t.device)
        # for VP: -1/2 beta(1 - t)(score(x, t) - x).
        # Note that this is - ~f(x(1-t), 1-t) in the SDE paper
        score = self.network(t, theta_t, *context_data)
        beta = self.beta(1 - t)
        return -1 / 2 * beta[:, None] * (score - theta_t)

    def get_likelihood_weighting(self, weighting):
        if weighting == "score-matching":

            def weighting(t):
                return self.sigma(t) ** 2

        elif weighting == "score-flow":

            def weighting(t):
                return self.beta(1 - t)

        else:
            raise NotImplementedError("invalid weighting function")
        return weighting

    def alpha(self, t):
        T = t * self.beta_min + 1 / 2 * (self.beta_max - self.beta_min) * (t**2)
        return torch.exp(-1 / 2 * T)

    def beta(self, t):
        return self.beta_min + t * (self.beta_max - self.beta_min)

    def mu(self, t, x_1):
        return self.alpha(1 - t)[:, None].to(x_1.device) * x_1

    def sigma(self, t):
        return torch.sqrt(1 - self.alpha(1 - t) ** 2)

    def get_t_theta_t_score(self, theta_1):
        # sample time in [0, 1 - epsilon]
        t = self.sample_t(len(theta_1))
        theta_0 = self.sample_theta_0(len(theta_1))
        mu, sigma = self.mu(t, theta_1), self.sigma(t)
        theta_t = mu + theta_0 * sigma[:, None]
        score = theta_0 / sigma[:, None]
        return t, theta_t, score



import time
import os
import numpy as np
from os.path import join, isfile
import csv
from typing import Literal


class AvgTracker:
    def __init__(self):
        self.sum = 0
        self.N = 0
        self.x = None

    def update(self, x, n=1):
        self.sum += x
        self.N += n
        self.x = x

    def get_avg(self):
        if self.N == 0:
            return float("nan")
        return self.sum / self.N


class EarlyStopping:
    """
    Implement early stopping during training, once the validation loss stops decreasing
    for a certain number of epochs (the patience).

    If val_loss > min_val_loss - delta for more than patience epochs, then returns
    early stopping occurs.
    """

    def __init__(
        self,
        patience: int = 5,
        verbose: bool = False,
        delta: float = 0.0,
        metric: Literal["training", "validation"] = "validation",
    ):
        """
        Parameters
        ----------
        patience: int = 5
            Number of epochs to wait before stopping.
        verbose: bool = False
            Whether to print counter increments.
        delta: float = 0.0
            Amount by which loss must decrease in patience epochs.
        metric: Literal["training", "validation"]
            Whether to use the training loss to determine early stopping ("training") or the test loss ("validation")
        """
        self.patience = patience
        self.verbose = verbose
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.val_loss_min = np.Inf
        self.delta = delta
        if metric not in ["training", "validation"]:
            raise ValueError(
                "Early Stopping metric must be 'training' or 'validation'."
            )
        self.metric = metric

    def __call__(self, val_loss: float):
        """
        Parameters
        ----------
        val_loss: float
            Value of the validation loss.
        model

        Returns
        -------
        bool
            Whether the current model has the lowest validation loss so-far.
        """
        score = -val_loss

        if self.best_score is None:
            self.best_score = score
            return True
        elif score < self.best_score + self.delta:
            self.counter += 1
            if self.verbose:
                print(f"EarlyStopping counter: {self.counter} out of {self.patience}")
            if self.counter >= self.patience:
                self.early_stop = True
            return False
        else:
            self.best_score = score
            self.counter = 0
            return True


class LossInfo:
    def __init__(self, epoch, len_dataset, batch_size, mode="Train", print_freq=1):
        # data for print statements
        self.epoch = epoch
        self.len_dataset = len_dataset
        self.batch_size = batch_size
        self.mode = mode
        self.print_freq = print_freq
        # track loss
        self.loss_tracker = AvgTracker()
        self.loss = None
        # track computation times
        self.times = {"Dataloader": AvgTracker(), "Network": AvgTracker()}
        self.t = time.time()

    def update_timer(self, timer_mode="Dataloader"):
        self.times[timer_mode].update(time.time() - self.t)
        self.t = time.time()

    def update(self, loss, n):
        self.loss = loss
        self.loss_tracker.update(loss * n, n)
        self.update_timer(timer_mode="Network")

    def get_avg(self):
        return self.loss_tracker.get_avg()

    def print_info(self, batch_idx):
        if batch_idx % self.print_freq == 0:
            print(
                "{} Epoch: {} [{}/{} ({:.0f}%)]".format(
                    self.mode,
                    self.epoch,
                    min(batch_idx * self.batch_size, self.len_dataset),
                    self.len_dataset,
                    100.0 * batch_idx * self.batch_size / self.len_dataset,
                ),
                end="\t\t",
            )
            # print loss
            print(f"Loss: {self.loss:.3f} ({self.get_avg():.3f})", end="\t\t")
            # print computation times
            td, td_avg = self.times["Dataloader"].x, self.times["Dataloader"].get_avg()
            tn, tn_avg = self.times["Network"].x, self.times["Network"].get_avg()
            print(f"Time Dataloader: {td:.3f} ({td_avg:.3f})", end="\t\t")
            print(f"Time Network: {tn:.3f} ({tn_avg:.3f})")


class RuntimeLimits:
    """
    Keeps track of the runtime limits (time limit, epoch limit, max. number
    of epochs for model).
    """

    def __init__(
        self,
        max_time_per_run: float = None,
        max_epochs_per_run: int = None,
        max_epochs_total: int = None,
        epoch_start: int = None,
    ):
        """

        Parameters
        ----------
        max_time_per_run: float = None
            maximum time for run, in seconds
            [soft limit, break only after full epoch]
        max_epochs_per_run: int = None
            maximum number of epochs for run
        max_epochs_total: int = None
            maximum total number of epochs for model
        epoch_start: int = None
            start epoch of run
        """
        self.max_time_per_run = max_time_per_run
        self.max_epochs_per_run = max_epochs_per_run
        self.max_epochs_total = max_epochs_total
        self.epoch_start = epoch_start
        self.time_start = time.time()
        if max_epochs_per_run is not None and epoch_start is None:
            raise ValueError("epoch_start required to check " "max_epochs_per_run.")

    def limits_exceeded(self, epoch: int = None):
        """
        Check whether any of the runtime limits are exceeded.

        Parameters
        ----------
        epoch: int = None

        Returns
        -------
        limits_exceeded: bool
            flag whether runtime limits are exceeded and run should be stopped;
            if limits_exceeded = True, this prints a message for the reason
        """
        # check time limit for run
        if self.max_time_per_run is not None:
            if time.time() - self.time_start >= self.max_time_per_run:
                print(
                    f"Stop run: Time limit of {self.max_time_per_run} s " f"exceeded."
                )
                return True
        # check epoch limit for run
        if self.max_epochs_per_run is not None:
            if epoch is None:
                raise ValueError("epoch required")
            if epoch - self.epoch_start >= self.max_epochs_per_run:
                print(
                    f"Stop run: Epoch limit of {self.max_epochs_per_run} per run reached."
                )
                return True
        # check total epoch limit
        if self.max_epochs_total is not None:
            if epoch >= self.max_epochs_total:
                print(
                    f"Stop run: Total epoch limit of {self.max_epochs_total} reached."
                )
                return True
        # return False if none of the limits is exceeded
        return False

    def local_limits_exceeded(self, epoch: int = None):
        """
        Check whether any of the local runtime limits are exceeded. Local runtime
        limits include max_epochs_per_run and max_time_per_run, but not max_epochs_total.

        Parameters
        ----------
        epoch: int = None

        Returns
        -------
        limits_exceeded: bool
            flag whether local runtime limits are exceeded
        """
        # check time limit for run
        if self.max_time_per_run is not None:
            if time.time() - self.time_start >= self.max_time_per_run:
                return True
        # check epoch limit for run
        if self.max_epochs_per_run is not None:
            if epoch is None:
                raise ValueError("epoch required")
            if epoch - self.epoch_start >= self.max_epochs_per_run:
                return True
        # return False if none of the limits is exceeded
        return False


def write_history(
    log_dir,
    epoch,
    train_loss,
    test_loss,
    learning_rates,
    aux=None,
    filename="history.txt",
):
    """
    Writes losses and learning rate history to csv file.

    Parameters
    ----------
    log_dir: str
        directory containing the history file
    epoch: int
        epoch
    train_loss: float
        train_loss of epoch
    test_loss: float
        test_loss of epoch
    learning_rates: list
        list of learning rates in epoch
    aux: list = []
        list of auxiliary information to be logged
    filename: str = 'history.txt'
        name of history file
    """
    if aux is None:
        aux = []
    history_file = join(log_dir, filename)
    if epoch == 1:
        assert not isfile(
            history_file
        ), f"File {history_file} exists, aborting to not overwrite it."

    with open(history_file, "w" if epoch == 1 else "a") as f:
        writer = csv.writer(f, delimiter="\t")
        writer.writerow([epoch, train_loss, test_loss, *learning_rates, *aux])


def copyfile(src, dst):
    """
    copy src to dst.
    :param src:
    :param dst:
    :return:
    """
    os.system("cp -p %s %s" % (src, dst))


def save_model(pm, log_dir, model_prefix="model", checkpoint_epochs=None):
    """
    Save model to <model_prefix>_latest.pt in log_dir. Additionally,
    all checkpoint_epochs a permanent checkpoint is saved.

    Parameters
    ----------
    pm:
        model to be saved
    log_dir: str
        log directory, where model is saved
    model_prefix: str = 'model'
        prefix for name of save model
    checkpoint_epochs: int = None
        number of steps between two consecutive model checkpoints
    """
    # save current model
    model_name = join(log_dir, f"{model_prefix}_latest.pt")
    print(f"Saving model to {model_name}.", end=" ")
    pm.save_model(model_name, save_training_info=True)
    print("Done.")

    # potentially copy model to a checkpoint
    if checkpoint_epochs is not None and pm.epoch % checkpoint_epochs == 0:
        model_name_cp = join(log_dir, f"{model_prefix}_{pm.epoch:03d}.pt")
        print(f"Copy model to checkpoint {model_name_cp}.", end=" ")
        copyfile(model_name, model_name_cp)
        print("Done.")



from dingo.core.posterior_models.base_model import BasePosteriorModel
from dingo.core.posterior_models.flow_matching import FlowMatchingPosteriorModel
from dingo.core.posterior_models.normalizing_flow import NormalizingFlowPosteriorModel
from dingo.core.posterior_models.score_matching import ScoreDiffusionPosteriorModel
from dingo.core.utils.backward_compatibility import (
    torch_load_with_fallback,
    update_model_config,
    check_minimum_version,
)


def build_model_from_kwargs(
    filename: str = None, settings: dict = None, **kwargs
) -> BasePosteriorModel:
    """
    Returns a PosteriorModel based on a saved network or settings dict.

    The function is careful to choose the appropriate PosteriorModel class (e.g.,
    for a normalizing flow, flow matching, or score matching).

    Parameters
    ----------
    filename: str
        Path to a saved network (.pt).
    settings: dict
        Settings dictionary.
    kwargs
        Arguments forwarded to the model constructor.

    Returns
    -------
    PosteriorModel
    """
    if (filename is None) == (settings is None):
        raise ValueError(
            "Either a filename or a settings dict must be provided, but not both."
        )

    models_dict = {
        "normalizing_flow": NormalizingFlowPosteriorModel,
        "flow_matching": FlowMatchingPosteriorModel,
        "score_matching": ScoreDiffusionPosteriorModel,
    }

    if filename is not None:
        d, _ = torch_load_with_fallback(filename, preferred_map_location="meta")
        if "version" in d:
            check_minimum_version(d["version"])
        else:
            # version was introduced in v0.3.3
            check_minimum_version("dingo=0.3.2")
        update_model_config(d["metadata"]["train_settings"]["model"])  # Backward compat
        posterior_model_type = d["metadata"]["train_settings"]["model"][
            "posterior_model_type"
        ]
    else:
        update_model_config(settings["train_settings"]["model"])  # Backward compat
        posterior_model_type = settings["train_settings"]["model"][
            "posterior_model_type"
        ]

    if not posterior_model_type.lower() in models_dict:
        raise ValueError("No valid posterior model type specified.")

    model = models_dict[posterior_model_type.lower()]

    return model(model_filename=filename, metadata=settings, **kwargs)


def autocomplete_model_kwargs(model_kwargs: dict, data_sample: list):
    """
    Autocomplete the model kwargs from train_settings and data_sample from the dataloader:

    * set input dimension of embedding net to shape of data_sample[1]
    * set dimension of parameter space to len(data_sample[0])
    * set added_context flag of embedding net if required for gnpe proxies
    * set context dim of posterior model to output dim of embedding net + gnpe proxy dim

    Parameters
    ----------
    model_kwargs: dict
        Model settings, which are modified in-place.
    data_sample: list
        Sample from dataloader (e.g., wfd[0]) used for autocomplection.
        Should be of format [parameters, GW data, gnpe_proxies], where the
        last element is only there is GNPE proxies are required.
    """

    # set input dims from ifo_list and domain information
    model_kwargs["embedding_kwargs"]["input_dims"] = list(data_sample[1].shape)
    # set dimension of parameter space of posterior model
    model_kwargs["posterior_kwargs"]["input_dim"] = len(data_sample[0])
    # set added_context flag of embedding net if GNPE proxies are required
    # set context dim of nsf to output dim of embedding net + GNPE proxy dim
    try:
        gnpe_proxy_dim = len(data_sample[2])
        model_kwargs["embedding_kwargs"]["added_context"] = True
        model_kwargs["posterior_kwargs"]["context_dim"] = (
            model_kwargs["embedding_kwargs"]["output_dim"] + gnpe_proxy_dim
        )
    except IndexError:
        model_kwargs["embedding_kwargs"]["added_context"] = False
        model_kwargs["posterior_kwargs"]["context_dim"] = model_kwargs[
            "embedding_kwargs"
        ]["output_dim"]



class GetItem:

    def __init__(self, key):
        self.key = key

    def __call__(self, sample):
        return sample[self.key]

class RenameKey:

    def __init__(self, old, new):
        self.old = old
        self.new = new

    def __call__(self, input_sample : dict):
        sample = input_sample.copy()
        sample[self.new] = sample.pop(self.old)
        return sample



import os

from bilby_pipe.job_creation.nodes import GenerationNode as BilbyGenerationNode
from bilby_pipe.utils import BilbyPipeError

from dingo.pipe.utils import _strip_unwanted_submission_keys


class GenerationNode(BilbyGenerationNode):

    def __init__(self, inputs, trigger_time, idx, dag, parent=None, importance_sampling=False):
        """
        Node for data generation jobs with DINGO-specific customizations
        """
        self.importance_sampling = importance_sampling
        
        # Call Node.__init__ directly, skipping BilbyGenerationNode.__init__
        super(BilbyGenerationNode, self).__init__(inputs, retry=3)
        
        # Copy all the BilbyGenerationNode logic but add DINGO customizations
        if not inputs.osg and inputs.generation_pool == "igwn-pool":
            raise BilbyPipeError(
                "Generation job requested to use the igwn-pool "
                "(OSG, --generation-pool=igwn-pool), but --osg=False"
            )
        else:
            self.run_node_on_osg = inputs.generation_pool == "igwn-pool"
        self.inputs = inputs
        self.trigger_time = trigger_time
        self.inputs.trigger_time = trigger_time
        self.idx = idx
        self.dag = dag
        self.request_cpus = 1

        self.setup_arguments()
        self.arguments.add("label", self.label)
        self.arguments.add("idx", self.idx)
        self.arguments.add("trigger-time", self.trigger_time)
        if self.inputs.injection_file is not None:
            self.arguments.add("injection-file", self.inputs.injection_file)
        if self.inputs.timeslide_file is not None:
            self.arguments.add("timeslide-file", self.inputs.timeslide_file)

        frame_files, success = self.resolve_frame_files
        need_scitokens = not success

        if self.inputs.transfer_files or self.inputs.osg:
            input_files_to_transfer = list()
            for attr in [
                "complete_ini_file",
                "prior_file",
                "injection_file",
                "gps_file",
                "timeslide_file",
            ]:
                if (value := getattr(self.inputs, attr)) is not None:
                    input_files_to_transfer.append(os.path.abspath(str(value)))

            # modifying the paths for OSDF networks
            network_files = []
            for s in [self.inputs.model, self.inputs.model_init]:
                if s is None:
                    continue
                if "osdf" in s:
                    # stripping osdf prefix as it is not needed
                    network_files.append(f"igwn+osdf://{s.replace('/osdf', '')}")
                else:
                    network_files.append(s)

            input_files_to_transfer.extend(network_files)

            if self.transfer_container:
                input_files_to_transfer.append(self.inputs.container)
            for value in [
                self.inputs.psd_dict,
                self.inputs.spline_calibration_envelope_dict,
                frame_files,
            ]:
                input_files_to_transfer.extend(self.extract_paths_from_dict(value))
            input_files_to_transfer.extend(self.inputs.additional_transfer_paths)

            input_files_to_transfer, need_auth = self.job_needs_authentication(
                input_files_to_transfer
            )

            # Credentials are needed to access any OSDF files
            if any(["osdf" in s for s in input_files_to_transfer]):
                need_auth = True
            need_scitokens = need_scitokens or need_auth

            self.extra_lines.extend(
                self._condor_file_transfer_lines(
                    input_files_to_transfer,
                    [self._relative_topdir(self.inputs.outdir, self.inputs.initialdir)],
                )
            )
            self.arguments.add("outdir", os.path.relpath(self.inputs.outdir))

        elif new_frames := [
            fname
            for fname in self.extract_paths_from_dict(frame_files)
            if fname.startswith(self.inputs.data_find_urltype)
        ]:
            from bilby_pipe.utils import logger
            logger.warning(
                "The following frame files were identified by gwdatafind for this analysis. "
                "These frames may not be found by the data generation stage as file "
                "transfer is not being used. You should either set transfer-files=True or "
                "pass these frame files to the data-dict option. You may need to "
                f"remove a prefix, e.g., file://localhost.\n\t{new_frames}"
            )
        if need_scitokens:
            self.extra_lines.extend(self.scitoken_lines)

        # DINGO-specific customization: Add site selection
        if self.inputs.generation_pool == "igwn-pool":
            sites = getattr(self.inputs, 'generation_desired_sites', None)
            if sites is not None:
                self.extra_lines.append(f'MY.DESIRED_Sites = "{sites}"')

        self.process_node()
        if parent:
            self.job.add_parent(parent.job)

        if self.inputs.simple_submission:
            _strip_unwanted_submission_keys(self.job)

    @property
    def executable(self):
        return self._get_executable_path("dingo_pipe_generation")

    def setup_arguments(self, **kwargs):
        super().setup_arguments(**kwargs)
        if self.importance_sampling:
            self.arguments.add_flag("importance-sampling-generation")

    @property
    def job_name(self):
        flag = "_IS" if self.importance_sampling else ""
        return super().job_name + flag

    @property
    def event_data_file(self):
        return os.path.join(
            self.inputs.data_directory, "_".join([self.label, "event_data.hdf5"])
        )



import numpy as np
import lal
import lalsimulation as LS


def linked_list_modes_to_dict_modes(hlm_ll):
    """Convert linked list of modes into dictionary with keys (l,m)."""
    hlm_dict = {}

    mode = hlm_ll.this
    while mode is not None:
        l, m = mode.l, mode.m
        hlm_dict[(l, m)] = mode.mode
        mode = mode.next

    return hlm_dict


def get_tapering_window_for_complex_time_series(h, tapering_flag: int = 1):
    """
    Get window for tapering of a complex time series from the lal backend. This is done
    by  tapering the time series with lal, and dividing tapered output by untapered
    input. lal does not support tapering of complex time series objects, so as a
    workaround we taper only the real part of the array and extract the window based on
    this.

    Parameters
    ----------
    h:
        complex lal time series object
    tapering_flag: int = 1
        Flag for tapering. See e.g. lines 2773-2777 in
            https://lscsoft.docs.ligo.org/lalsuite/lalsimulation/
            _l_a_l_sim_inspiral_waveform_taper_8c_source.html#l00222
        tapering_flag = 1 corresponds to LAL_SIM_INSPIRAL_TAPER_START

    Returns
    -------
    window: np.ndarray
        Array of length h.data.length, with the window used for tapering.
    """
    h_tapered = lal.CreateREAL8TimeSeries(
        "h_tapered", h.epoch, 0, h.deltaT, None, h.data.length
    )
    h_tapered.data.data = h.data.data.copy().real
    LS.SimInspiralREAL8WaveTaper(h_tapered.data, tapering_flag)
    eps = 1e-20 * np.max(np.abs(h.data.data))
    window = (np.abs(h_tapered.data.data) + eps) / (np.abs(h.data.data.real) + eps)
    # FIXME: using eps for numerical stability is not really robust here
    return window


def taper_td_modes_in_place(hlm_td, tapering_flag: int = 1):
    """
    Taper the time domain modes in place.

    Parameters
    ----------
    hlm_td: dict
        Dictionary with (l,m) keys and the complex lal time series objects for the
        corresponding modes.
    tapering_flag: int = 1
        Flag for tapering. See e.g. lines 2773-2777 in
            https://lscsoft.docs.ligo.org/lalsuite/lalsimulation/
            _l_a_l_sim_inspiral_waveform_taper_8c_source.html#l00222
        tapering_flag = 1 corresponds to LAL_SIM_INSPIRAL_TAPER_START
    """
    for _, h in hlm_td.items():
        window = get_tapering_window_for_complex_time_series(h, tapering_flag)
        h.data.data *= window


def td_modes_to_fd_modes(hlm_td, domain):
    """
    Transform dict of td modes to dict of fd modes via FFT. The td modes are expected
    to be tapered.

    Parameters
    ----------
    hlm_td: dict
        Dictionary with (l,m) keys and the complex lal time series objects for the
        corresponding tapered modes.
    domain: dingo.gw.domains.UniformFrequencyDomain
        Target domain after FFT.

    Returns
    -------
    hlm_fd: dict
        Dictionary with (l,m) keys and numpy arrays with the corresponding modes as
        values.
    """
    hlm_fd = {}

    delta_f = domain.delta_f
    delta_t = 0.5 / domain.f_max
    f_nyquist = domain.f_max  # use f_max as f_nyquist
    chirplen = int(2 * f_nyquist / delta_f)
    # sample frequencies, -f_max,...,-f_min,...0,...,f_min,...,f_max
    freqs = np.concatenate((-domain()[::-1], domain()[1:]), axis=0)
    # For even chirplength, we get chirplen + 1 output frequencies. However, the f_max
    # and -f_max bins are redundant, so we have chirplen unique bins.
    assert len(freqs) == chirplen + 1

    lal_fft_plan = lal.CreateForwardCOMPLEX16FFTPlan(chirplen, 0)
    for lm, h_td in hlm_td.items():
        assert np.abs(h_td.deltaT - delta_t) < 1e-12

        # resize data to chirplen by zero-padding or truncating
        # if chirplen < h_td.data.length:
        #     print(
        #         f"Specified frequency interval of {delta_f} Hz is too large "
        #         f"for a chirp of duration {h_td.data.length * delta_t} s with "
        #         f"Nyquist frequency {f_nyquist} Hz. The inspiral will be "
        #         f"truncated."
        #     )
        lal.ResizeCOMPLEX16TimeSeries(h_td, h_td.data.length - chirplen, chirplen)

        # Initialize a lal frequency series. We choose length chirplen + 1, while h_td is
        # only of length chirplen. This means, that the last bin h_fd.data.data[-1]
        # will not be modified by the lal FFT, and we have to copy over h_fd.data.data[0]
        # to h_fd.data.data[-1]. This corresponds to setting h(-f_max) = h(f_max).
        h_fd = lal.CreateCOMPLEX16FrequencySeries(
            "h_fd", h_td.epoch, 0, delta_f, None, chirplen + 1
        )
        # apply FFT
        lal.COMPLEX16TimeFreqFFT(h_fd, h_td, lal_fft_plan)
        assert np.abs(h_fd.deltaF - delta_f) < 1e-10
        assert np.abs(h_fd.f0 + domain.f_max) < 1e-6

        # time shift
        dt = (
            1.0 / h_fd.deltaF + h_fd.epoch.gpsSeconds + h_fd.epoch.gpsNanoSeconds * 1e-9
        )
        hlm_fd[lm] = h_fd.data.data * np.exp(-1j * 2 * np.pi * dt * freqs)
        # Set h(-f_max) = h(f_max), see above
        hlm_fd[lm][-1] = hlm_fd[lm][0]

    return hlm_fd


def get_polarizations_from_fd_modes_m(hlm_fd, iota, phase):
    pol_m = {}
    polarizations = ["h_plus", "h_cross"]

    for (l, m), h in hlm_fd.items():
        if m not in pol_m:
            pol_m[m] = {k: 0.0 for k in polarizations}
            pol_m[-m] = {k: 0.0 for k in polarizations}

        # In the L0 frame, we compute the polarizations from the modes using the
        # spherical harmonics below.
        ylm = lal.SpinWeightedSphericalHarmonic(iota, np.pi / 2 - phase, -2, l, m)
        ylmstar = ylm.conjugate()

        # Modes (l,m) are defined on domain -f_max,...,-f_min,...0,...,f_min,...,f_max.
        # This splits up the frequency series into positive and negative frequency parts.
        if len(h) % 2 != 1:
            raise ValueError(
                "Even number of bins encountered, should be odd: -f_max,...,0,...,f_max."
            )
        offset = len(h) // 2
        h1 = h[offset:]
        h2 = h[offset::-1].conj()

        # Organize the modes such that pol_m[m] transforms as e^{- 1j * m * phase}.
        # This differs from the usual way, e.g.,
        #   https://lscsoft.docs.ligo.org/lalsuite/lalsimulation/
        #   _l_a_l_sim_inspiral_8c_source.html#l04801
        pol_m[m]["h_plus"] += 0.5 * h1 * ylm
        pol_m[-m]["h_plus"] += 0.5 * h2 * ylmstar
        pol_m[m]["h_cross"] += 0.5 * 1j * h1 * ylm
        pol_m[-m]["h_cross"] += -0.5 * 1j * h2 * ylmstar

    return pol_m


def get_starting_frequency_for_SEOBRNRv5_conditioning(parameters):
    """
    Compute starting frequency needed for having 3 extra cycles for tapering the TD modes.
    It returns the needed quantities to apply the standard LALSimulation conditioning routines to the TD modes.

    Parameters
    ----------
    parameters: dict
        Dictionary of parameters suited for GWSignal (obtained with NewInterfaceWaveformGenerator._convert_parameters)

    Returns
    ----------
    f_min: float
      Waveform starting frequency
    f_start: float
      New waveform starting frequency
    extra_time: float
      Extra time to take care of situations where the frequency is close to merger
    original_f_min: float
      Initial waveform starting frequency
    f_isco: float
      ISCO frequency
    """

    extra_time_fraction = (
        0.1  # fraction of waveform duration to add as extra time for tapering
    )
    extra_cycles = 3.0  # more extra time measured in cycles at the starting frequency

    f_min = parameters["f22_start"].value
    m1 = parameters["mass1"].value
    m2 = parameters["mass2"].value
    S1z = parameters["spin1z"].value
    S2z = parameters["spin2z"].value
    original_f_min = f_min

    f_isco = 1.0 / (pow(9.0, 1.5) * np.pi * (m1 + m2) * lal.MTSUN_SI)
    if f_min > f_isco:
        f_min = f_isco

    # upper bound on the chirp time starting at f_min
    tchirp = LS.SimInspiralChirpTimeBound(
        f_min, m1 * lal.MSUN_SI, m2 * lal.MSUN_SI, S1z, S2z
    )
    # upper bound on the final black hole spin */
    spinkerr = LS.SimInspiralFinalBlackHoleSpinBound(S1z, S2z)
    # upper bound on the final plunge, merger, and ringdown time */
    tmerge = LS.SimInspiralMergeTimeBound(
        m1 * lal.MSUN_SI, m2 * lal.MSUN_SI
    ) + LS.SimInspiralRingdownTimeBound((m1 + m2) * lal.MSUN_SI, spinkerr)

    # extra time to include for all waveforms to take care of situations where the frequency is close to merger (and is sweeping rapidly): this is a few cycles at the low frequency
    textra = extra_cycles / f_min
    # compute a new lower frequency
    f_start = LS.SimInspiralChirpStartFrequencyBound(
        (1.0 + extra_time_fraction) * tchirp + tmerge + textra,
        m1 * lal.MSUN_SI,
        m2 * lal.MSUN_SI,
    )

    f_isco = 1.0 / (pow(6.0, 1.5) * np.pi * (m1 + m2) * lal.MTSUN_SI)

    return f_min, f_start, extra_time_fraction * tchirp + textra, original_f_min, f_isco


def taper_td_modes_for_SEOBRNRv5_extra_time(
    h, extra_time, f_min, original_f_min, f_isco
):
    """
    Apply standard tapering procedure mimicking LALSimulation routine (https://lscsoft.docs.ligo.org/lalsuite/lalsimulation/_l_a_l_sim_inspiral_generator_conditioning_8c.html#ac78b5fcdabf8922a3ac479da20185c85)

    Parameters
    ----------
    h:
        complex gwpy TimeSeries object
    extra_time: float
        Extra time to take care of situations where the frequency is close to merger
    f_min: float
        Starting frequency employed in waveform generation
    original_f_min: float
        Initial starting frequency requested by the user
    f_isco:
        ISCO frequency

    Returns
    ----------
    h_return
        complex lal timeseries object
    """

    # Split in real and imaginary parts, since LAL conditioning routines are for real timeseries
    h_tapered_re = lal.CreateREAL8TimeSeries(
        "h_tapered", h.epoch.value, 0, h.dt.value, None, len(h)
    )
    h_tapered_re.data.data = h.value.copy().real

    h_tapered_im = lal.CreateREAL8TimeSeries(
        "h_tapered_im", h.epoch.value, 0, h.dt.value, None, len(h)
    )
    h_tapered_im.data.data = h.value.copy().imag

    # condition the time domain waveform by tapering in the extra time at the beginning and high-pass filtering above original f_min
    LS.SimInspiralTDConditionStage1(
        h_tapered_re, h_tapered_im, extra_time, original_f_min
    )
    # final tapering at the beginning and at the end to remove filter transients
    # waveform should terminate at a frequency >= Schwarzschild ISCO
    # so taper one cycle at this frequency at the end; should not make
    # any difference to IMR waveforms */
    LS.SimInspiralTDConditionStage2(h_tapered_re, h_tapered_im, f_min, f_isco)

    # Construct complex timeseries
    h_return = lal.CreateCOMPLEX16TimeSeries(
        "h_return",
        h_tapered_re.epoch,
        0,
        h_tapered_re.deltaT,
        None,
        h_tapered_re.data.length,
    )

    h_return.data.data = h_tapered_re.data.data + 1j * h_tapered_im.data.data

    # return timeseries
    return h_return



import numpy as np
import scipy
import pandas as pd


def generate_cornerplot(*sample_sets, filename=None):
    try:
        from chainconsumer import ChainConsumer
    except ImportError:
        return

    parameters = [
        p
        for p in sample_sets[0]["samples"].keys()
        if p in set.intersection(*tuple(set(s["samples"].keys()) for s in sample_sets))
    ]
    N = len(sample_sets)

    c = ChainConsumer()
    for s in sample_sets:
        c.add_chain(s["samples"][parameters], color=s["color"], name=s["name"])
    c.configure(
        linestyles=["-"] * N,
        linewidths=[1.5] * N,
        sigmas=[np.sqrt(2) * scipy.special.erfinv(x) for x in [0.5, 0.9]],
        shade=[False] + [True] * (N - 1),
        shade_alpha=0.3,
        bar_shade=False,
        label_font_size=10,
        tick_font_size=10,
        usetex=False,
        legend_kwargs={"fontsize": 30},
        kde=0.7,
    )
    c.plotter.plot(filename=filename)


def load_ref_samples(ref_samples_file, drop_geocent_time=True):
    # Todo: this function should be made more flexible for other formats and parameters
    from bilby.gw.conversion import component_masses_to_chirp_mass

    columns = [
        "mass_1",
        "mass_2",
        "phase",
        "geocent_time",
        "luminosity_distance",
        "a_1",
        "a_2",
        "tilt_1",
        "tilt_2",
        "phi_12",
        "phi_jl",
        "theta_jn",
        "psi",
        "ra",
        "dec",
    ]
    samples = np.load(ref_samples_file, allow_pickle=True)["samples"][:,:15]
    samples = pd.DataFrame(data=samples, columns=columns)
    # add chirp mass and mass ratio.
    Mc = component_masses_to_chirp_mass(samples["mass_1"], samples["mass_2"])
    q = samples["mass_2"] / samples["mass_1"]
    samples.insert(loc=0, column="chirp_mass", value=Mc)
    samples.insert(loc=0, column="mass_ratio", value=q)
    if drop_geocent_time:
        samples.drop(columns="geocent_time", inplace=True)
    return samples



from multiprocessing import Pool

import numpy as np
import pandas as pd
from threadpoolctl import threadpool_limits


class Likelihood(object):
    def log_likelihood(self, theta):
        raise NotImplementedError("log_likelihood() should be implemented in subclass.")

    def log_likelihood_multi(
        self, theta: pd.DataFrame, num_processes: int = 1
    ) -> np.ndarray:
        """
        Calculate the log likelihood at multiple points in parameter space. Works with
        multiprocessing.

        This wraps the log_likelihood() method.

        Parameters
        ----------
        theta : pd.DataFrame
            Parameters values at which to evaluate likelihood.
        num_processes : int
            Number of processes to use.

        Returns
        -------
        np.array of log likelihoods
        """
        with threadpool_limits(limits=1, user_api="blas"):

            # Generator object for theta rows. For idx this yields row idx of
            # theta dataframe, converted to dict, ready to be passed to
            # self.log_likelihood.
            theta_generator = (d[1].to_dict() for d in theta.iterrows())

            if num_processes > 1:
                with Pool(processes=num_processes) as pool:
                    log_likelihood = pool.map(self.log_likelihood, theta_generator)
            else:
                log_likelihood = list(map(self.log_likelihood, theta_generator))

        return np.array(log_likelihood)



#
#  Adapted from bilby_pipe. In particular, uses the bilby_pipe data generation code.
#
import copy
import json
import os

import numpy as np

import dingo.pipe.create_injections

from bilby.core.prior import PriorDict
from bilby_pipe.input import Input
from bilby_pipe.main import MainInput as BilbyMainInput
from bilby_pipe.utils import (
    convert_string_to_dict,
    get_colored_string,
    get_command_line_arguments,
    logger,
    parse_args,
    convert_prior_string_input,
    BilbyPipeError,
)

from .dag_creator import generate_dag
from .parser import create_parser
from .utils import dict_to_string

from ..gw.domains.build_domain import build_domain_from_model_metadata
from dingo.core.posterior_models.build_model import build_model_from_kwargs
from ..gw.inference.gw_samplers import check_frequency_updates
from ..gw.injection import Injection
from ..gw.noise.asd_dataset import ASDDataset

logger.name = "dingo_pipe"


def fill_in_arguments_from_model(args, perform_arg_checks=True):
    if perform_arg_checks:
        if args.prior_dict is not None:
            raise ValueError(
                "Do not specify prior-dict in INI file. This is obtained from "
                "the DINGO model. To update the prior, specify "
                "prior-dict-updates."
            )
        if args.model_reference_time is not None:
            raise ValueError(
                "Do not specify model-reference-time in INI file. This is obtained from the "
                "DINGO model."
            )

    logger.info(f"Loading dingo model from {args.model} in order to access settings.")

    try:
        model = build_model_from_kwargs(
            filename=args.model, device="meta", load_training_info=False
        )
    except RuntimeError:
        # 'meta' is not supported by older version of python / torch
        model = build_model_from_kwargs(
            filename=args.model, device=args.device, load_training_info=False
        )

    model_metadata = model.metadata

    domain = build_domain_from_model_metadata(model_metadata, base=True)

    prior = Injection.from_posterior_model_metadata(model_metadata).prior
    geocent_time_prior = prior["geocent_time"]
    deltaT = geocent_time_prior.maximum - geocent_time_prior.minimum

    # Offset needed in case of DeltaFunction priors not peaked at 0.0.
    Toffset = 0.5 * (geocent_time_prior.minimum + geocent_time_prior.maximum)

    # Dingo and Bilby have different conventions for the geocent_time prior. Dingo
    # always centers it around something close to 0.0, whereas Bilby centers it around
    # the trigger time. Drop the geocent_time parameter here so that bilby_pipe
    # re-builds it as needed for Bilby.
    del prior["geocent_time"]

    data_settings = model_metadata["train_settings"]["data"]

    # In dingo_pipe, we download and prepare data based on the model frequency range.
    # This is because different maximum frequencies for different detectors would
    # require separate domain objects within Dingo, and this is not implemented.
    # Instead, we update frequency ranges using masking within Dingo. Updated frequency
    # ranges are passed around within Dingo using the event_metadata.

    model_args = {
        "duration": domain.duration,
        "minimum_frequency": domain.f_min,
        "maximum_frequency": domain.f_max,
        "detectors": data_settings["detectors"],
        "waveform_approximant": model_metadata["dataset_settings"][
            "waveform_generator"
        ]["approximant"],
        "reference_frequency": model_metadata["dataset_settings"]["waveform_generator"][
            "f_ref"
        ],
        "deltaT": deltaT,
        "Toffset": Toffset,
        "prior_dict": dict_to_string(prior),
        "model_reference_time": model_metadata["train_settings"]["data"]["ref_time"],
    }

    changed_args = {}
    for k, v in model_args.items():
        args_v = getattr(args, k)
        if args_v is not None:
            # Convert type from str to enable comparison.
            try:
                # it's possible that the attribute is a dict
                # or list.
                # for example, the minimum-frequency could be
                # {H1: 20, L1:20}
                if isinstance(args_v, str) and "{" and "}" in args_v:
                    args_v = convert_string_to_dict(args_v)
                elif isinstance(args_v, list):
                    # if it's a list we need to strip quotes
                    # e.g. detectors = ["'H1'", "'L1'"]
                    # should become ["H1", "L1"] for the comparison
                    args_v = [l.strip("'") for l in args_v]
                elif isinstance(v, float):
                    args_v = float(args_v)
                elif isinstance(v, int):
                    args_v = int(args_v)
            except ValueError:
                pass

            if args_v != v:
                # this is for when the minimum-frequency is a float
                # but args_v is a dict
                if isinstance(args_v, dict) and all(
                    val == v for val in args_v.values()
                ):
                    continue

                if k in ["waveform_approximant"]:
                    raise NotImplementedError(
                        "Cannot change waveform approximant during importance sampling."
                    )  # TODO: Implement this. Also no error if passed explicitly as an update.
                if k in ["minimum_frequency", "maximum_frequency"]:
                    logger.info(
                        f"Strain-cropping: Plan to update network {k} from {v} to"
                        f" {args_v}."
                    )
                    continue  # Do not update args to model value.
                logger.warning(
                    f"Argument {k} provided to dingo_pipe as {args_v} "
                    f"inconsistent with value {v} in model file. Using model value for "
                    f"inference, and will attempt to change this during importance "
                    f"sampling."
                )
                changed_args[k] = args_v

        setattr(args, k, v)

    frequency_input = Input([], [], print_msg=False)
    frequency_input.detectors = model_args["detectors"]
    frequency_input.sampling_frequency = args.sampling_frequency
    frequency_input.minimum_frequency = args.minimum_frequency
    frequency_input.maximum_frequency = args.maximum_frequency
    check_frequency_updates(
        model_metadata,
        frequency_input.minimum_frequency_dict,
        frequency_input.maximum_frequency_dict,
    )

    # TODO: Also check consistency between model and init_model settings.

    # If an ASDDataset is specified, pull out PSDs and save them as .txt files. Use
    # these for downstream tasks by specifying the psd_dict.
    if args.asd_dataset:
        asd_dataset = ASDDataset(file_name=args.asd_dataset)
        domain_dict = copy.deepcopy(domain.domain_dict)
        asd_dataset.update_domain(domain_dict)
        psd_dict = {}
        rng = np.random.default_rng(args.generation_seed)
        for ifo_name in args.detectors:
            psd_path = asd_dataset.save_psd(args.outdir, ifo_name, rng=rng)
            psd_dict[ifo_name] = str(psd_path)
        args.asd_dataset = None
        args.psd_dict = str(psd_dict)

    # Updates that are explicitly provided take priority.
    if args.importance_sampling_updates is None:
        importance_sampling_updates = {}
    else:
        importance_sampling_updates = convert_string_to_dict(
            args.importance_sampling_updates
        )
        importance_sampling_updates = {
            k.replace("-", "_"): v for k, v in importance_sampling_updates.items()
        }
    return {**changed_args, **importance_sampling_updates}, model_args


class MainInput(BilbyMainInput):
    def __init__(
        self, args, unknown_args, importance_sampling_updates, perform_checks=True
    ):
        # Settings added for dingo.

        self.model = args.model
        self.model_init = args.model_init
        self.num_gnpe_iterations = args.num_gnpe_iterations
        self.importance_sampling_updates = importance_sampling_updates
        self.prior_dict_updates = args.prior_dict_updates
        self.model_reference_time = args.model_reference_time

        Input.__init__(self, args, unknown_args, print_msg=False)

        self.known_args = args
        self.unknown_args = unknown_args
        self.ini = args.ini
        self.submit = args.submit
        self.condor_job_priority = args.condor_job_priority
        self.create_summary = args.create_summary
        self.scitoken_issuer = args.scitoken_issuer
        self.container = args.container

        # self.cosmology = args.cosmology

        self.outdir = args.outdir
        self.label = args.label
        self.log_directory = args.log_directory
        self.accounting = args.accounting
        self.accounting_user = args.accounting_user
        # self.sampler = args.sampler
        self.detectors = args.detectors
        self.coherence_test = (
            False  # dingo mod: Cannot use different sets of detectors.
        )
        self.data_dict = args.data_dict
        self.channel_dict = args.channel_dict
        self.frame_type_dict = args.frame_type_dict
        self.data_find_url = args.data_find_url
        self.data_find_urltype = args.data_find_urltype
        self.n_parallel = args.n_parallel
        # useful when condor nodes don't have access to submit filesystem
        self.transfer_files = args.transfer_files
        self.additional_transfer_paths = args.additional_transfer_paths
        self.osg = args.osg
        self.desired_sites = (
            args.cpu_desired_sites
        )  # Dummy variable so bilby_pipe doesn't complain.
        self.cpu_desired_sites = args.cpu_desired_sites
        self.gpu_desired_sites = args.gpu_desired_sites
        self.generation_desired_sites = args.generation_desired_sites
        # self.analysis_executable = args.analysis_executable
        # self.analysis_executable_parser = args.analysis_executable_parser
        self.result_format = "hdf5"
        self.final_result = args.final_result
        self.final_result_nsamples = args.final_result_nsamples

        self.webdir = args.webdir
        self.email = args.email
        self.notification = args.notification
        self.queue = args.queue
        self.existing_dir = args.existing_dir

        self.scheduler = args.scheduler
        self.scheduler_args = args.scheduler_args
        self.scheduler_module = args.scheduler_module
        self.scheduler_env = args.scheduler_env
        self.scheduler_analysis_time = args.scheduler_analysis_time
        self.disable_hdf5_locking = args.disable_hdf5_locking
        self.environment_variables = args.environment_variables
        self.getenv = args.getenv

        self.waveform_approximant = args.waveform_approximant
        #
        # self.time_reference = args.time_reference
        self.time_reference = "geocent"
        # self.reference_frame = args.reference_frame
        # self.likelihood_type = args.likelihood_type
        self.duration = args.duration
        # self.phase_marginalization = args.phase_marginalization
        self.prior_file = None  # Dingo update. To change prior use the prior_dict.
        self.prior_dict = args.prior_dict
        self.default_prior = "BBHPriorDict"
        self.minimum_frequency = args.minimum_frequency
        self.enforce_signal_duration = args.enforce_signal_duration

        self.run_local = args.local
        self.generation_pool = args.generation_pool
        self.local_generation = args.local_generation
        self.local_plot = args.local_plot

        self.post_trigger_duration = args.post_trigger_duration

        # self.ignore_gwpy_data_quality_check = args.ignore_gwpy_data_quality_check
        self.trigger_time = args.trigger_time
        self.deltaT = args.deltaT
        self.Toffset = args.Toffset
        self.gps_tuple = args.gps_tuple
        self.gps_file = args.gps_file
        self.timeslide_file = args.timeslide_file
        self.gaussian_noise = args.gaussian_noise
        self.zero_noise = args.zero_noise
        self.n_simulation = args.n_simulation

        self.injection = args.injection
        self.injection_numbers = args.injection_numbers
        self.injection_file = args.injection_file
        self.injection_dict = args.injection_dict
        self.save_bilby_data_dump = args.save_bilby_data_dump
        # self.injection_waveform_arguments = args.injection_waveform_arguments
        # self.injection_waveform_approximant = args.injection_waveform_approximant
        # self.injection_frequency_domain_source_model = (
        #     args.injection_frequency_domain_source_model
        # )
        # self.injection_waveform_generator_class_ctor_args = (
        #     args.injection_waveform_generator_constructor_dict
        # )
        self.generation_seed = args.generation_seed
        if self.importance_sampling_updates and self.gaussian_noise:
            raise ValueError(
                "Cannot update data for importance sampling if using "
                "simulated Gaussian noise. This risks inconsistent noise "
                "realizations."
            )

        self.importance_sample = args.importance_sample

        self.request_disk = args.request_disk
        self.request_memory_generation = args.request_memory_generation
        self.request_memory = args.request_memory
        self.request_memory_importance_sampling = (
            args.request_memory_importance_sampling
        )
        self.request_cpus = args.request_cpus
        self.request_cpus_importance_sampling = args.request_cpus_importance_sampling
        # self.sampler_kwargs = args.sampler_kwargs
        # self.mpi_samplers = ["pymultinest"]
        # self.use_mpi = (self.sampler in self.mpi_samplers) and (self.request_cpus > 1)
        #
        # # Set plotting options when need the plot node
        self.plot_node_needed = False
        for plot_attr in [
            "corner",
            "weights",
            "log_probs",
        ]:
            attr = f"plot_{plot_attr}"
            setattr(self, attr, getattr(args, attr))
            if getattr(self, attr):
                self.plot_node_needed = True

        self.plot_pp = args.plot_pp
        #
        # # Set all other plotting options
        # for plot_attr in [
        #     "trace",
        #     "data",
        #     "injection",
        #     "spectrogram",
        #     "format",
        # ]:
        #     attr = f"plot_{plot_attr}"
        #     setattr(self, attr, getattr(args, attr))
        #
        # self.postprocessing_executable = args.postprocessing_executable
        # self.postprocessing_arguments = args.postprocessing_arguments
        # self.single_postprocessing_executable = args.single_postprocessing_executable
        # self.single_postprocessing_arguments = args.single_postprocessing_arguments
        #
        self.summarypages_arguments = args.summarypages_arguments

        self.psd_dict = args.psd_dict
        self.psd_maximum_duration = args.psd_maximum_duration
        self.psd_length = args.psd_length
        self.psd_fractional_overlap = args.psd_fractional_overlap
        self.psd_start_time = args.psd_start_time
        self.spline_calibration_envelope_dict = args.spline_calibration_envelope_dict

        if perform_checks:
            # self.check_source_model(args)
            # self.check_calibration_prior_boundary(args)
            # self.check_cpu_parallelisation()
            self.check_accounting()
            if self.injection:
                self.check_injection()

        self.requirements = []
        self.device = args.device
        self.simple_submission = args.simple_submission

        if args.extra_lines:
            self.extra_lines = args.extra_lines
        else:
            self.extra_lines = []

        if args.sampling_requirements:
            self.sampling_requirements = args.sampling_requirements
        else:
            self.sampling_requirements = []

    @property
    def request_cpus_importance_sampling(self):
        return self._request_cpus_importance_sampling

    @request_cpus_importance_sampling.setter
    def request_cpus_importance_sampling(self, request_cpus_importance_sampling):
        logger.info(
            f"Setting analysis request_cpus_importance_sampling = "
            f"{request_cpus_importance_sampling}"
        )
        self._request_cpus_importance_sampling = request_cpus_importance_sampling

    @property
    def prior_dict_updates(self):
        """The input prior_dict from the ini (if given)

        Note, this is not the bilby prior (see self.priors for that), this is
        a key-val dictionary where the val's are strings which are converting
        into bilby priors in `_get_prior
        """
        return self._prior_dict_updates

    @prior_dict_updates.setter
    def prior_dict_updates(self, prior_dict_updates):
        if isinstance(prior_dict_updates, dict):
            prior_dict_updates = prior_dict_updates
        elif isinstance(prior_dict_updates, str):
            prior_dict_updates = convert_prior_string_input(prior_dict_updates)
        elif prior_dict_updates is None:
            self._prior_dict_updates = None
            return
        else:
            raise BilbyPipeError(
                f"prior_dict_updates={prior_dict_updates} not " f"understood"
            )

        self._prior_dict_updates = {
            self._convert_prior_dict_key(key): val
            for key, val in prior_dict_updates.items()
        }

    def _get_priors(self, add_time=True):
        priors = super()._get_priors(add_time=add_time)
        priors.update(PriorDict(self.prior_dict_updates))
        return priors

    def check_accounting(self):
        if self.accounting is None:
            logger.warning(
                "No accounting tag specified. If using IGWN computing resources"
                " please refer to https://computing.docs.ligo.org/guide/htcondor/accounting/"
            )
        else:
            try:
                purpose = self.accounting.split(".")[1]
                if purpose == "prod":
                    logger.info(
                        get_colored_string(
                            "Using a 'prod' accounting tag: please ensure this is intentional"
                        )
                    )
                pipeline = self.accounting.split(".")[-1]
                if pipeline != "dingo":
                    logger.info(
                        get_colored_string(
                            "Using a non-'dingo' accounting tag: please ensure this is "
                            "intentional"
                        )
                    )
            except ValueError:
                logger.debug(
                    "Accounting tag does not follow the IGWN format. If you are using non-IGWN"
                    " resources this is fine, otherwise please refer to"
                    " https://computing.docs.ligo.org/guide/htcondor/accounting/"
                )
        logger.info(f"Using accounting tag: {self.accounting}")
        logger.info(f"Using accounting-user: {self.accounting_user}")

    def check_injection(self):
        """Check injection behaviour

        If injections are requested, either use the injection-dict,
        injection-file, or create an injection-file

        """
        default_injection_file_name = "{}/{}_injection_file.dat".format(
            self.data_directory, self.label
        )
        if self.injection_dict is not None:
            logger.debug(
                "Using injection dict from ini file {}".format(
                    json.dumps(self.injection_dict, indent=2)
                )
            )
        elif self.injection_file is not None:
            logger.debug(f"Using injection file {self.injection_file}")
        elif os.path.isfile(default_injection_file_name):
            # This is done to avoid overwriting the injection file
            logger.debug(f"Using injection file {default_injection_file_name}")
            self.injection_file = default_injection_file_name
        else:
            logger.debug("No injection file found, generating one now")

            if self.gps_file is not None or self.gps_tuple is not None:
                if self.n_simulation > 0 and self.n_simulation != len(self.gpstimes):
                    raise BilbyPipeError(
                        "gps_file/gps_tuple option and n_simulation are not matched"
                    )
                gpstimes = self.gpstimes
                n_injection = len(gpstimes)
            else:
                gpstimes = None
                n_injection = self.n_simulation

            if self.trigger_time is None:
                # Use the model reference time as the trigger time by default. This
                # avoids needing to adjust sky position in post-postprocessing,
                # and it avoids complications due to DeltaFunction ra priors.
                trigger_time_injections = self.model_reference_time
            else:
                trigger_time_injections = self.trigger_time

            dingo.pipe.create_injections.create_injection_file(
                filename=default_injection_file_name,
                prior_file=self.prior_file,
                prior_dict=self._get_priors(
                    add_time=False
                ),  # ensure prior-dict-updates incorporated
                n_injection=n_injection,
                trigger_time=trigger_time_injections,
                deltaT=self.deltaT,
                Toffset=self.Toffset,  # dingo-pipe mod
                gpstimes=gpstimes,
                duration=self.duration,
                post_trigger_duration=self.post_trigger_duration,
                generation_seed=self.generation_seed,
                extension="dat",
                default_prior=self.default_prior,
            )
            self.injection_file = default_injection_file_name

        # Check the gps_file has the sample length as number of simulation
        if self.gps_file is not None:
            if len(self.gpstimes) != len(self.injection_df):
                raise BilbyPipeError("Injection file length does not match gps_file")

        if self.n_simulation > 0:
            if self.n_simulation != len(self.injection_df):
                raise BilbyPipeError(
                    "n-simulation does not match the number of injections: "
                    "please check your ini file"
                )
        elif self.n_simulation == 0 and self.gps_file is None:
            self.n_simulation = len(self.injection_df)
            logger.debug(
                f"Setting n_simulation={self.n_simulation} to match injections"
            )


def write_complete_config_file(parser, args, inputs, input_cls=MainInput):
    args_dict = vars(args).copy()
    for key, val in args_dict.items():
        if key == "label":
            continue
        if isinstance(val, str):
            if os.path.isfile(val) or os.path.isdir(val):
                if not args.osg:
                    setattr(args, key, os.path.abspath(val))
                else:
                    setattr(args, key, val)
        if isinstance(val, list):
            if isinstance(val[0], str):
                setattr(args, key, f"[{', '.join(val)}]")
    # args.sampler_kwargs = str(inputs.sampler_kwargs)
    args.importance_sampling_updates = str(inputs.importance_sampling_updates)
    args.submit = False
    parser.write_to_file(
        filename=inputs.complete_ini_file,
        args=args,
        overwrite=False,
        include_description=False,
    )

    # Verify that the written complete config is identical to the source config
    # complete_args = parser.parse([inputs.complete_ini_file])
    # complete_inputs = input_cls(complete_args, "")
    # ignore_keys = ["scheduler_module", "submit"]
    # differences = []
    # for key, val in inputs.__dict__.items():
    #     if key in ignore_keys:
    #         continue
    #     if key not in complete_args:
    #         continue
    #     if isinstance(val, pd.DataFrame) and all(val == complete_inputs.__dict__[key]):
    #         continue
    #     if isinstance(val, np.ndarray) and all(
    #         np.array(val) == np.array(complete_inputs.__dict__[key])
    #     ):
    #         continue
    #     if isinstance(val, str) and os.path.isfile(val):
    #         # Check if it is relpath vs abspath
    #         if os.path.abspath(val) == os.path.abspath(complete_inputs.__dict__[key]):
    #             continue
    #     if val == complete_inputs.__dict__[key]:
    #         continue
    #
    #     differences.append(key)
    #
    # if len(differences) > 0:
    #     for key in differences:
    #         print(key, f"{inputs.__dict__[key]} -> {complete_inputs.__dict__[key]}")
    # raise BilbyPipeError(
    #     "The written config file {} differs from the source {} in {}".format(
    #         inputs.ini, inputs.complete_ini_file, differences
    #     )
    # )


def main():
    parser = create_parser(top_level=True)
    args, unknown_args = parse_args(get_command_line_arguments(), parser)

    importance_sampling_updates, model_args = fill_in_arguments_from_model(args)
    inputs = MainInput(args, unknown_args, importance_sampling_updates)
    write_complete_config_file(parser, args, inputs)

    # TODO: Use two sets of inputs! The first must match the network; the second is
    #  used in importance sampling.

    generate_dag(inputs)

    if len(unknown_args) > 0:
        msg = get_colored_string(f"Unrecognized arguments {unknown_args}")
        logger.warning(msg)



from abc import ABC, abstractmethod
from dingo.gw.gwutils import *


class Domain(ABC):
    """Defines the physical domain on which the data of interest live.

    This includes a specification of the bins or points,
    and a few additional properties associated with the data.
    """

    @abstractmethod
    def __len__(self) -> int:
        """Number of bins or points in the domain"""
        pass

    @abstractmethod
    def __call__(self, *args, **kwargs) -> np.ndarray:
        """Array of bins in the domain"""
        pass

    @abstractmethod
    def update(self, new_settings: dict):
        pass

    @abstractmethod
    def time_translate_data(self, data, dt) -> np.ndarray:
        """Time translate strain data by dt seconds."""
        pass

    @property
    @abstractmethod
    def noise_std(self) -> float:
        """Standard deviation of the whitened noise distribution"""
        # FIXME: For this to make sense, it assumes knowledge about how the domain is used in conjunction
        #  with (waveform) data, whitening and adding noise. Is this the best place to define this?
        pass

    @property
    @abstractmethod
    def sampling_rate(self) -> float:
        """The sampling rate of the data [Hz]."""
        pass

    @property
    @abstractmethod
    def f_max(self) -> float:
        """The maximum frequency [Hz] is set to half the sampling rate."""

    @property
    @abstractmethod
    def duration(self) -> float:
        """Waveform duration in seconds."""
        pass

    @property
    @abstractmethod
    def min_idx(self) -> int:
        pass

    @property
    @abstractmethod
    def max_idx(self) -> int:
        pass

    @property
    @abstractmethod
    def domain_dict(self):
        """Enables to rebuild the domain via calling build_domain(domain_dict)."""
        pass

    def __eq__(self, other):
        if self.domain_dict == other.domain_dict:
            return True
        else:
            return False



import os
import tempfile
import uuid
from pathlib import Path
from typing import Generator
import yaml

import numpy as np
import pytest
import bilby

from dingo.gw.dataset.generate_dataset import _generate_dataset_main
from dingo.gw.dataset.waveform_dataset import WaveformDataset
from dingo.gw.domains import Domain
from dingo.gw.dataset import generate_dataset

SETTINGS_YAML_SMALL = """\
# settings for domain of waveforms
domain:
  type: UniformFrequencyDomain
  f_min: 10.0
  f_max: 1024.0
  delta_f: 1.0

# settings for waveform generator
waveform_generator:
  approximant: IMRPhenomPv2
  f_ref: 20.0

# settings for intrinsic prior over parameters
intrinsic_prior:
  # prior for non-fixed parameters
  mass_1: bilby.core.prior.Constraint(minimum=10.0, maximum=80.0, name='mass_1')
  mass_2: bilby.core.prior.Constraint(minimum=10.0, maximum=80.0, name='mass_2')
  mass_ratio: bilby.core.prior.Uniform(minimum=0.125, maximum=1.0, name='mass_ratio')
  chirp_mass: bilby.core.prior.Uniform(minimum=25.0, maximum=100.0, name='chirp_mass')
  phase: default
  a_1: bilby.core.prior.Uniform(minimum=0.0, maximum=0.88, name='a_1')
  a_2: bilby.core.prior.Uniform(minimum=0.0, maximum=0.88, name='a_2')
  tilt_1: default
  tilt_2: default
  phi_12: default
  phi_jl: default
  theta_jn: default
  # reference values for fixed (extrinsic) parameters
  luminosity_distance: 100.0 # Mpc
  geocent_time: 0.0 # s

num_samples: 50

compression:
  svd:
    num_training_samples: 10
    size: 5
"""


@pytest.fixture(scope="session")
def temp_dir() -> Generator[Path, None, None]:
    """
    Fixture to provide a temporary directory path using pathlib.Path.

    Returns
    -------
    A pathlib.Path object pointing to the temporary directory.
    """
    with tempfile.TemporaryDirectory() as tmpdirname:
        yield Path(tmpdirname)


@pytest.fixture(scope="session")
def generate_waveform_dataset_small(temp_dir: Path) -> Path:
    """
    Create a small waveform dataset on-the-fly
    in a temporary directory.
    """

    # Create temp directory and settings file
    path = temp_dir / "tmp_test" / str(uuid.uuid4())
    path.mkdir(parents=True)
    settings_path = path / "settings.yaml"
    out_file = path / "waveform_dataset.hdf5"
    num_processes = 4

    with open(settings_path, "w") as fp:
        fp.writelines(SETTINGS_YAML_SMALL)

    _generate_dataset_main(str(settings_path), str(out_file), num_processes)

    return path


@pytest.mark.slow
def test_load_waveform_dataset(generate_waveform_dataset_small):
    wfd_path = generate_waveform_dataset_small

    path = f"{wfd_path}/waveform_dataset.hdf5"
    wfd = WaveformDataset(file_name=path, precision="single")

    assert len(wfd) > 0
    el = wfd[0]
    assert isinstance(el, dict)
    assert set(el.keys()) == {"parameters", "waveform"}
    assert isinstance(el["parameters"], dict)
    assert isinstance(el["waveform"], dict)
    assert isinstance(el["waveform"]["h_plus"], np.ndarray)
    assert isinstance(el["waveform"]["h_cross"], np.ndarray)

    # Check the associated domain
    assert len(wfd.domain) > 0
    assert isinstance(wfd.domain, Domain)
    assert isinstance(wfd.domain.domain_dict, dict)

    # Check the associated data settings
    data_settings_keys = {
        "domain",
        "waveform_generator",
        "intrinsic_prior",
        "num_samples",
        "compression",
    }
    assert set(wfd.settings.keys()) == data_settings_keys

    """Check that using a list of indices also works."""
    ind_list = [0, 1, 2]
    el_list = wfd.__getitems__(ind_list)
    assert isinstance(el_list, list)
    assert len(el_list) == len(ind_list)
    assert set(el_list[0].keys()) == {"parameters", "waveform"}
    assert isinstance(el_list[0]["parameters"], dict)
    assert isinstance(el_list[0]["waveform"], dict)
    assert isinstance(el_list[0]["waveform"]["h_plus"], np.ndarray)
    assert isinstance(el_list[0]["waveform"]["h_cross"], np.ndarray)


@pytest.mark.slow
def test_truncation_of_waveform(generate_waveform_dataset_small):
    """Check truncation of wfd."""
    wfd_path = generate_waveform_dataset_small

    path = f"{wfd_path}/waveform_dataset.hdf5"
    wfd = WaveformDataset(file_name=path, precision="single")
    el = wfd[0]

    f_min = wfd.domain.f_min
    f_max = wfd.domain.f_max
    delta_f = wfd.domain._delta_f

    # check that truncation works as intended when setting new range
    f_min_new = 20
    f_max_new = 100
    wfd2 = WaveformDataset(
        file_name=path, domain_update={"f_min": f_min_new, "f_max": f_max_new}
    )
    assert len(wfd2.domain) == len(wfd2.domain())
    # check that new domain settings are correctly adapted
    assert wfd2.domain.f_min == f_min_new
    assert wfd2.domain.f_max == f_max_new
    assert wfd2.domain._delta_f == wfd.domain._delta_f
    # check that truncation works as intended
    for pol in ["h_cross", "h_plus"]:
        # f_min_new to f_max_new check
        a = el["waveform"][pol][int(f_min_new / delta_f) : int(f_max_new / delta_f) + 1]
        b = wfd2[0]["waveform"][pol][int(f_min_new / delta_f) :]
        scale_factor = np.max(np.abs(a))
        assert len(a) == f_max_new / delta_f + 1 - f_min_new / delta_f
        assert np.allclose(b / scale_factor, a / scale_factor)
        assert not np.allclose(b / scale_factor, np.roll(a, 1) / scale_factor)

        # f_min to f_min_new check
        a = el["waveform"][pol][int(f_min / delta_f) : int(f_min_new / delta_f)]
        b = wfd2[0]["waveform"][pol][int(f_min / delta_f) : int(f_min_new / delta_f)]
        assert not np.allclose(b / scale_factor, a / scale_factor)

        # below f_min_new check
        assert np.all(wfd2[0]["waveform"][pol][: int(f_min_new)] == 0.0)
    assert len(wfd2.domain) == f_max_new / delta_f + 1
    assert len(wfd2.domain) == len(wfd2.domain())


@pytest.mark.slow
def test_load_waveform_dataset_with_leave_polarizations_on_disk(
    generate_waveform_dataset_small,
):
    wfd_path = generate_waveform_dataset_small

    path = f"{wfd_path}/waveform_dataset.hdf5"
    wfd = WaveformDataset(
        file_name=path,
        precision="single",
        leave_waveforms_on_disk=True,
    )

    assert len(wfd) > 0
    el = wfd[0]
    assert isinstance(el, dict)
    assert set(el.keys()) == {"parameters", "waveform"}
    assert isinstance(el["parameters"], dict)
    assert isinstance(el["waveform"], dict)
    assert isinstance(el["waveform"]["h_plus"], np.ndarray)
    assert isinstance(el["waveform"]["h_cross"], np.ndarray)

    """Check that using a list of indices also works."""
    ind_list = [0, 1, 2, 3, 4, 5]
    el_list = wfd.__getitems__(ind_list)
    assert isinstance(el_list, list)
    assert len(el_list) == len(ind_list)
    assert set(el_list[0].keys()) == {"parameters", "waveform"}
    assert isinstance(el_list[0]["parameters"], dict)
    assert isinstance(el_list[0]["waveform"], dict)
    assert isinstance(el_list[0]["waveform"]["h_plus"], np.ndarray)
    assert isinstance(el_list[0]["waveform"]["h_cross"], np.ndarray)

    """Check that using a randomly ordered list of indices provides the expected ordering."""
    ind_list_rand = np.random.permutation(ind_list).tolist()
    el_list_rand = wfd.__getitems__(ind_list_rand)
    np.testing.assert_equal([el_list[i] for i in ind_list_rand], el_list_rand)

    """Check that waveforms and parameters are loaded from disk equivalently to loading the full dataset."""
    wfd_full = WaveformDataset(
        file_name=path,
        precision="single",
        leave_waveforms_on_disk=False,
    )
    el_list_full = wfd_full.__getitems__(ind_list)
    np.testing.assert_equal(el_list, el_list_full)


@pytest.fixture
def wfd_settings():
    settings = yaml.safe_load(SETTINGS_YAML_SMALL)
    return settings


class BinaryPrior(bilby.prior.Prior):
    def __init__(self):
        super().__init__(name="failing_prior")

    def sample(self, size):
        return np.random.choice([30, -1], size=size, p=[0.8, 0.2])


@pytest.fixture
def binary_prior():
    return BinaryPrior()


def test_wfd_size(wfd_settings: str, binary_prior: BinaryPrior):
    """
    Test that the size requested by the waveform generator settings is the same as the
    size of the generated dataset. This should be the case even when there are failures
    in the waveform generation
    """
    # changing the waveform generator settings to create a prior which will create
    # failing waveforms for a fraction of the prior. I.e. can't generate negative
    # chirp masses so the waveform generator will fail
    wfd_settings["intrinsic_prior"]["chirp_mass"] = binary_prior
    del wfd_settings["intrinsic_prior"]["mass_1"]
    del wfd_settings["intrinsic_prior"]["mass_2"]
    del wfd_settings["compression"]
    wfd = generate_dataset(wfd_settings, 1)
    assert len(wfd) == wfd_settings["num_samples"]



import argparse
import ast
import os

import h5py
import torch
import yaml

from dingo.core.utils.backward_compatibility import torch_load_with_fallback


def parse_args():

    parser = argparse.ArgumentParser()
    parser.add_argument("--checkpoint", type=str, required=True)
    parser.add_argument("--settings_file", type=str, required=True)
    parser.add_argument("--out_file", type=str, required=True)
    parser.add_argument("--train_dir", type=str, required=True)
    return parser.parse_args()


def main():
    args = parse_args()

    with open(args.settings_file, "r") as f:
        train_settings = yaml.safe_load(f)

    d, _ = torch_load_with_fallback(args.checkpoint)

    data = {
        "waveform_dataset_path": train_settings["waveform_dataset_path"],
        "train_fraction": train_settings["train_settings"]["train_fraction"],
        "conditioning": train_settings["data_conditioning"],
    }
    data.update(train_settings["transform_settings"])

    model = d["model_kwargs"]
    model["type"] = "nsf+embedding"
    model["embedding_net_kwargs"]["svd"] = (
        {
            "size": model["embedding_net_kwargs"]["n_rb"],
        },
    )
    del model["embedding_net_kwargs"]["n_rb"]
    del model["embedding_net_kwargs"]["V_rb_list"]

    training = {
        "stage_0": {
            "epochs": train_settings["train_settings"]["runtime_limits"][
                "max_epochs_total"
            ],
            "asd_dataset_path": train_settings["asd_dataset_path"],
            "freeze_rb_layer": train_settings["train_settings"]["freeze_rb_layer"],
            "optimizer": train_settings["train_settings"]["optimizer_kwargs"],
            "scheduler": train_settings["train_settings"]["scheduler_kwargs"],
            "batch_size": train_settings["train_settings"]["batch_size"],
        }
    }

    local = {
        "device": train_settings["train_settings"]["device"],
        "num_workers": train_settings["train_settings"]["num_workers"],
        "runtime_limits": train_settings["train_settings"]["runtime_limits"],
        "checkpoint_epochs": train_settings["train_settings"]["checkpoint_epochs"],
        "condor": train_settings["condor_settings"],
    }
    del local["runtime_limits"]["max_epochs_total"]

    d["metadata"] = {
        "train_settings": {
            "data": data,
            "model": model,
            "training": training,
        }
    }

    # Save (for posterity) the waveform dataset settings
    f = h5py.File(d["metadata"]["train_settings"]["data"]["waveform_dataset_path"], "r")
    settings = ast.literal_eval(f.attrs["settings"])
    d["metadata"]["dataset_settings"] = settings
    f.close()

    torch.save(d, args.out_file)

    # Save local settings
    with open(os.path.join(args.train_dir, "local_settings.yaml"), "w") as f:
        yaml.dump(local, f, default_flow_style=False, sort_keys=False)


if __name__ == "__main__":
    main()



import copy
from pathlib import Path
from typing import Iterable, Optional

import numpy as np

from dingo.gw.domains import build_domain, UniformFrequencyDomain
from dingo.gw.domains.base_frequency_domain import BaseFrequencyDomain
from dingo.gw.gwutils import *
from dingo.gw.dataset import DingoDataset

HIGH_ASD_VALUE = 1.0


class ASDDataset(DingoDataset):
    """
    Dataset of amplitude spectral densities (ASDs). The ASDs are typically
    used for whitening strain data, and additionally passed as context to the
    neural density estimator.
    """

    dataset_type = "asd_dataset"

    def __init__(
        self,
        file_name=None,
        dictionary=None,
        ifos=None,
        precision=None,
        domain_update=None,
    ):
        """
        Parameters
        ----------
        file_name : str
            HDF5 file containing a dataset
        dictionary : dict
            Contains settings and data entries. The dictionary keys should be
            'settings', 'asds', and 'gps_times'.
        ifos : List[str]
            List of detectors used for dataset, e.g. ['H1', 'L1'].
            If not set, all available ones in the dataset are used.
        precision : str ('single', 'double')
            If provided, changes precision of loaded dataset.
        domain_update : dict
            If provided, update domain from existing domain using new settings.
        """
        self.precision = precision
        super().__init__(
            file_name=file_name,
            dictionary=dictionary,
            data_keys=["asds", "gps_times", "asd_parameterizations"],
        )

        if ifos is not None:
            for ifo in list(self.asds.keys()):
                if ifo not in ifos:
                    self.asds.pop(ifo)
                    self.gps_times.pop(ifo)

        if "window_factor" in self.settings["domain_dict"]:
            print(
                "Warning: 'window_factor' is no longer used in ASDDataset. "
                "Removing from settings."
            )
            self.settings["domain_dict"].pop("window_factor")
        self.domain = build_domain(self.settings["domain_dict"])
        if not check_domain_compatibility(self.asds, self.domain):
            raise ValueError("ASDs in dataset not compatible with domain.")
        if domain_update is not None:
            self.update_domain(domain_update)

        # Update dtypes if necessary
        if self.precision is not None:
            if self.precision == "single":
                for ifo, asd in self.asds.items():
                    self.asds[ifo] = asd.astype(np.float32, copy=False)
            elif self.precision == "double":
                for ifo, asd in self.asds.items():
                    self.asds[ifo] = asd.astype(np.float64, copy=False)
            else:
                raise TypeError(
                    'precision can only be changed to "single" or "double".'
                )

    @property
    def length_info(self):
        """The number of asd samples per detector."""
        return {key: len(val) for key, val in self.asds.items()}

    @property
    def gps_info(self):
        """Min/Max GPS time for each detector."""
        gps_info_dict = {}
        for key, val in self.gps_times.items():
            if not isinstance(val, Iterable):
                gps_info_dict[key] = val
            else:
                gps_info_dict[key] = (min(val), max(val))
        return gps_info_dict

    def update_domain(self, domain_update):
        """
        Update the domain based on new configuration. Also adjust data arrays to match
        the new domain.

        The ASD dataset provides ASDs in a particular domain. In Frequency domain,
        this is [0, domain.f_max]. In practice one may want to train a network based on
        slightly different domain settings, which corresponds to truncating the likelihood
        integral.

        This method provides functionality for that. It truncates the data below a
        new f_max, and sets the ASD below f_min to a large but finite value.

        Parameters
        ----------
        domain_update : dict
            Settings dictionary. Must contain a subset of the keys contained in
            domain_dict.
        """
        # Note that we require domain_update to have a type specified, even if
        # unchanged from the original domain. This reduces risks of errors.
        if self.domain.domain_dict["type"] == domain_update["type"]:
            self.domain.update(domain_update)
            self.settings["domain"] = copy.deepcopy(self.domain.domain_dict)

            # truncate the dataset
            for ifo, asds in self.asds.items():
                self.asds[ifo] = self.domain.update_data(
                    asds,
                    low_value=HIGH_ASD_VALUE,
                )

        elif (
            self.domain.domain_dict["type"] == "UniformFrequencyDomain"
            and domain_update["type"] == "MultibandedFrequencyDomain"
        ):
            print("Updating ASD dataset to MultibandedFrequencyDomain.")
            asd_dataset_decimated = {}
            mfd = build_domain(domain_update)
            ufd = mfd.base_domain
            if not check_domain_compatibility(self.asds, ufd):
                # If the ASD length is not compatible with the new base UFD,
                # first truncate it.
                print(
                    f"  Truncating first to new base UniformFrequencyDomain: f_max "
                    f"{self.domain.f_max} Hz -> {ufd.f_max} Hz"
                )
                self.domain.update(ufd.domain_dict)  # Additional compatibility check.
                for ifo, asds in self.asds.items():
                    self.asds[ifo] = self.domain.update_data(
                        asds,
                        low_value=HIGH_ASD_VALUE,
                    )

            decimation_method = "inverse-asd-decimation"

            for ifo, asds in self.asds.items():
                asd_dataset_decimated[ifo] = np.zeros((len(asds), len(mfd)))
                for idx, asd in enumerate(asds):
                    if decimation_method == "inverse-asd-decimation":
                        asd_dataset_decimated[ifo][idx, :] = 1 / mfd.decimate(1 / asd)
                    elif decimation_method == "psd-decimation":
                        asd_dataset_decimated[ifo][idx, :] = (
                            1e-20 * mfd.decimate((asd * 1e20) ** 2) ** 0.5
                        )
                    else:
                        raise NotImplementedError(
                            f"Unknown decimation method " f"{decimation_method}."
                        )

            self.asds = asd_dataset_decimated
            self.settings["domain_dict"] = mfd.domain_dict
            self.domain = mfd

        else:
            raise NotImplementedError(
                f"Cannot update ASD domain type "
                f"{self.domain.domain_dict['type']} to {domain_update['type']}"
            )

    def sample_random_asds(self, n: Optional[int] = None) -> dict[str, np.ndarray]:
        """
        Sample n random ASDs for each detector.

        Parameters
        ----------
        n : int
            Number of asds to sample

        Returns
        -------
        dict[str, np.ndarray]
            Where the keys correspond to the detectors and the values
            are arrays of shape (n, D) where D is the number of frequency bins
            and n is the number of ASDs requested. If n=None, then the
            function returns a single ASD for each detector, so the array is
            flattened to be shape D
        """
        if n is None:
            return {k: v[np.random.choice(len(v), 1)[0]] for k, v in self.asds.items()}
        else:
            return {k: v[np.random.choice(len(v), n)] for k, v in self.asds.items()}

    def save_psd(self, directory, ifo_name, idx: Optional[int] = None, rng=None):
        if rng is None:
            rng = np.random.default_rng()
        if idx is None:
            idx = rng.choice(len(self.asds[ifo_name]))
        directory = Path(directory)
        directory.mkdir(exist_ok=True)
        psd_path = directory / f"{ifo_name}_{idx}_psd.txt"
        np.savetxt(
            psd_path,
            np.vstack([self.domain(), self.asds[ifo_name][idx] ** 2]).T,
        )
        return psd_path


def check_domain_compatibility(data: dict, domain: BaseFrequencyDomain) -> bool:
    for v in data.values():
        if not domain.check_data_compatibility(v):
            return False
    return True



from typing import Union, Protocol

import numpy as np
import pandas as pd
from astropy.time import Time
from bilby.core.prior import PriorDict, DeltaFunction, Constraint
from bilby.gw.detector import InterferometerList
from torchvision.transforms import Compose

from dingo.core.samplers import Sampler, GNPESampler
from dingo.core.transforms import GetItem, RenameKey
from dingo.gw.domains import (
    MultibandedFrequencyDomain,
    build_domain_from_model_metadata,
    UniformFrequencyDomain,
    Domain,
)
from dingo.gw.domains import build_domain
from dingo.gw.gwutils import get_extrinsic_prior_dict
from dingo.gw.prior import build_prior_with_defaults
from dingo.gw.result import Result
from dingo.gw.transforms import (
    WhitenAndScaleStrain,
    RepackageStrainsAndASDS,
    ToTorch,
    SelectStandardizeRepackageParameters,
    GNPECoalescenceTimes,
    TimeShiftStrain,
    GNPEBase,
    PostCorrectGeocentTime,
    CopyToExtrinsicParameters,
    GetDetectorTimes,
    DecimateWaveformsAndASDS,
    MaskDataForFrequencyRangeUpdate,
)


class SamplerProtocol(Protocol):
    base_model_metadata: dict

    def _initialize_transforms(self) -> None:
        ...


class _GWMixinProtocol(SamplerProtocol):
    detectors: list[str]
    domain: Domain
    random_strain_cropping: dict


class GWSamplerMixin(object):
    """
    Mixin class designed to add gravitational wave functionality to Sampler classes:
        * builder for data domain
        * correction for fixed detector locations during training (t_ref)
    """

    def __init__(self: SamplerProtocol, **kwargs):
        """
        Parameters
        ----------
        kwargs
            Keyword arguments that are forwarded to the superclass.
        """
        # Has to be specified before init, because the information is required in _initialize_transforms()
        self._minimum_frequency = None
        self._maximum_frequency = None
        super().__init__(**kwargs)
        self.t_ref = self.base_model_metadata["train_settings"]["data"]["ref_time"]
        self._pesummary_package = "gw"
        self._result_class = Result

    @property
    def detectors(self: SamplerProtocol):
        return self.base_model_metadata["train_settings"]["data"]["detectors"]

    @property
    def random_strain_cropping(self: SamplerProtocol):
        return self.base_model_metadata["train_settings"]["data"].get(
            "random_strain_cropping"
        )

    @property
    def minimum_frequency(self) -> float | dict[str, float]:
        if self._minimum_frequency is not None:
            return self._minimum_frequency
        else:
            return self.domain.f_min

    @minimum_frequency.setter
    def minimum_frequency(self: _GWMixinProtocol, value: dict[str, float] | float):
        if isinstance(self.domain, MultibandedFrequencyDomain):
            domain = self.domain.base_domain
        elif isinstance(self.domain, UniformFrequencyDomain):
            domain = self.domain
        else:
            raise ValueError("Frequency updates only possible for frequency domains.")
        _validate_minimum_frequency(
            value,
            self.detectors,
            domain,
            self.random_strain_cropping,
        )  # TODO: Ensure minimum frequency is a dict?
        self._minimum_frequency = value
        self._initialize_transforms()

    @property
    def maximum_frequency(self) -> float | dict[str, float]:
        if self._maximum_frequency is not None:
            return self._maximum_frequency
        else:
            return self.domain.f_max

    @maximum_frequency.setter
    def maximum_frequency(self: _GWMixinProtocol, value: Union[float, dict]):
        if isinstance(self.domain, MultibandedFrequencyDomain):
            domain = self.domain.base_domain
        elif isinstance(self.domain, UniformFrequencyDomain):
            domain = self.domain
        else:
            raise ValueError("Frequency updates only possible for frequency domains.")
        _validate_maximum_frequency(
            value,
            self.detectors,
            domain,
            self.random_strain_cropping,
        )
        self._maximum_frequency = value
        self._initialize_transforms()

    @property
    def frequency_updates(self) -> bool:
        def normalize(val):
            if isinstance(val, dict):
                return set(val.values())
            return {val}

        return normalize(self.minimum_frequency) != {self.domain.f_min} or normalize(
            self.maximum_frequency
        ) != {self.domain.f_max}

    @property
    def event_metadata(self):
        if self._event_metadata is not None:
            metadata = self._event_metadata.copy()
        else:
            metadata = {}
        metadata["minimum_frequency"] = self.minimum_frequency
        metadata["maximum_frequency"] = self.maximum_frequency
        return metadata

    @event_metadata.setter
    def event_metadata(self, value):
        if value is not None:
            value = value.copy()
            if "minimum_frequency" in value:
                self.minimum_frequency = value.pop("minimum_frequency")
            if "maximum_frequency" in value:
                self.maximum_frequency = value.pop("maximum_frequency")
        self._event_metadata = value

    def _build_domain(self: Sampler):
        """
        Construct the domain object based on model metadata.

        Called by __init__() immediately after _build_prior().
        """
        self.domain = build_domain(
            self.base_model_metadata["dataset_settings"]["domain"]
        )

        data_settings = self.base_model_metadata["train_settings"]["data"]
        if "domain_update" in data_settings:
            self.domain.update(data_settings["domain_update"])


    def _correct_reference_time(
        self: Sampler, samples: Union[dict, pd.DataFrame], inverse: bool = False
    ):
        """
        Correct the sky position of an event based on the reference time of the model.
        This is necessary since the model was trained with with fixed detector (reference)
        positions. This transforms the right ascension based on the e difference between
        the time of the event and t_ref.

        The correction is only applied if the event time can be found in self.metadata[
        'event'].

        This method modifies the samples in place.

        Parameters
        ----------
        samples : dict or pd.DataFrame
        inverse : bool, default True
            Whether to apply instead the inverse transformation. This is used prior to
            calculating the log_prob.
        """
        if self.event_metadata is not None:
            t_event = self.event_metadata.get("time_event")
            if t_event is not None and t_event != self.t_ref and "ra" in samples:
                ra = samples["ra"]
                time_reference = Time(self.t_ref, format="gps", scale="utc")
                time_event = Time(t_event, format="gps", scale="utc")
                longitude_event = time_event.sidereal_time("apparent", "greenwich")
                longitude_reference = time_reference.sidereal_time(
                    "apparent", "greenwich"
                )
                delta_longitude = longitude_event - longitude_reference
                ra_correction = delta_longitude.rad
                if not inverse:
                    samples["ra"] = (ra + ra_correction) % (2 * np.pi)
                else:
                    samples["ra"] = (ra - ra_correction) % (2 * np.pi)

    def _post_process(self, samples: Union[dict, pd.DataFrame], inverse: bool = False):
        """
        Post-processing of parameter samples.
        * Add any fixed parameters from the prior.
        * Correct the sky position for a potentially fixed reference time.
          (see self._correct_reference_time)

        This method modifies the samples in place.

        Parameters
        ----------
        samples : dict or pd.DataFrame
        inverse : bool, default True
            Whether to apply instead the inverse transformation. This is used prior to
            calculating the log_prob.
        """
        intrinsic_prior = self.metadata["dataset_settings"]["intrinsic_prior"]
        extrinsic_prior = get_extrinsic_prior_dict(
            self.metadata["train_settings"]["data"]["extrinsic_prior"]
        )
        prior = build_prior_with_defaults({**intrinsic_prior, **extrinsic_prior})

        if not inverse:
            # Add fixed parameters from prior.
            num_samples = len(samples[list(samples.keys())[0]])
            for k, p in prior.items():
                if isinstance(p, DeltaFunction) and k not in samples:
                    v = p.peak
                    print(f"Adding fixed parameter {k} = {v} from prior.")
                    samples[k] = p.peak * np.ones(num_samples)
        else:
            # Drop non-inference parameters from samples.
            # NOTE: Important to drop "log_prob" in particular before running
            # Sampler.log_prob(), otherwise log probabilities are added.
            drop_parameters = [
                k for k in samples.keys() if k not in self.inference_parameters
            ]
            if isinstance(samples, pd.DataFrame):
                samples.drop(columns=drop_parameters, inplace=True, errors="ignore")
            elif isinstance(samples, dict):
                for k in drop_parameters:
                    samples.pop(k, None)

        if not self.unconditional_model:
            self._correct_reference_time(samples, inverse)


class GWSampler(GWSamplerMixin, Sampler):
    """
    Sampler for gravitational-wave inference using neural posterior estimation.
    Augments the base class by defining transform_pre and transform_post to prepare
    data for the inference network.

    transform_pre :
        * Decimates data (if necessary and using MultibandedFrequencyDomain).
        * Whitens strain.
        * Repackages strain data and the inverse ASDs (suitably scaled) into a torch
          tensor.

    transform_post :
        * Extract the desired inference parameters from the network output (
          array-like), de-standardize them, and repackage as a dict.

    Also mixes in GW functionality for building the domain and correcting the reference
    time.

    Allows for conditional and unconditional models, and draws samples from the model
    based on (optional) context data.

    This is intended for use either as a standalone sampler, or as a sampler producing
    initial sample points for a GNPE sampler.
    """

    def _initialize_transforms(self):
        # preprocessing transforms:
        transform_pre = []
        #   * in case of MultibandedFrequencyDomain, decimate data from base domain
        if isinstance(self.domain, MultibandedFrequencyDomain):
            transform_pre.append(
                DecimateWaveformsAndASDS(self.domain, decimation_mode="whitened")
            )

        #   * whiten and scale strain (since the inference network expects standardized
        #   data)
        transform_pre.append(WhitenAndScaleStrain(self.domain.noise_std))
        if self.frequency_updates:
            # * update frequency range
            # Needs to happen before RepackageStrainsAndASDs since we might need to apply
            # detectors specific frequency updates.
            transform_pre.append(
                MaskDataForFrequencyRangeUpdate(
                    domain=self.domain,
                    minimum_frequency=self.minimum_frequency,
                    maximum_frequency=self.maximum_frequency,
                )
            )
        #   * repackage strains and asds from dicts to an array
        #   * convert array to torch tensor on the correct device
        #   * extract only strain/waveform from the sample
        transform_pre += [
            # Use base metadata so that unconditional samplers still know how to
            # transform data, since this transform is used by the GNPE sampler as
            # well.
            RepackageStrainsAndASDS(
                ifos=self.detectors,
                first_index=self.domain.min_idx,
            ),
            ToTorch(device=self.model.device),
            GetItem("waveform"),
        ]
        self.transform_pre = Compose(transform_pre)

        # postprocessing transforms:
        #   * de-standardize data and extract inference parameters
        self.transform_post = SelectStandardizeRepackageParameters(
            {"inference_parameters": self.inference_parameters},
            self.metadata["train_settings"]["data"]["standardization"],
            inverse=True,
            as_type="dict",
        )


class GWSamplerGNPE(GWSamplerMixin, GNPESampler):
    """
    Gravitational-wave GNPE sampler. It wraps a PosteriorModel and a standard Sampler for
    initialization. The former is used to generate initial samples for Gibbs sampling.

    Compared to the base class, this class implements the required transforms for
    preparing data and parameters for the network. This includes GNPE transforms,
    data processing transforms, and standardization/de-standardization of parameters.

    A GNPE network is conditioned on additional "proxy" context theta^, i.e.,

    p(theta | theta^, d)

    The theta^ depend on theta via a fixed kernel p(theta^ | theta). Combining these
    known distributions, this class uses Gibbs sampling to draw samples from the joint
    distribution,

    p(theta, theta^ | d)

    The advantage of this approach is that we are allowed to perform any transformation of
    d that depends on theta^. In particular, we can use this freedom to simplify the
    data, e.g., by aligning data to have merger times = 0 in each detector. The merger
    times are unknown quantities that must be inferred jointly with all other
    parameters, and GNPE provides a means to do this iteratively. See
    https://arxiv.org/abs/2111.13139 for additional details.

    Gibbs sampling breaks access to the probability density, so this must be recovered
    through other means. One way is to train an unconditional flow to represent p(theta^
    | d) for fixed d based on the samples produced through the GNPE Gibbs sampling.
    Starting from these, a single Gibbs iteration gives theta from the GNPE network,
    along with the probability density in the joint space. This is implemented in
    GNPESampler provided the init_sampler provides proxies directly and num_iterations
    = 1.

    Attributes (beyond those of Sampler)
    ------------------------------------
    init_sampler : Sampler
        Used for providing initial samples for Gibbs sampling.
    num_iterations : int
        Number of Gibbs iterations to perform.
    iteration_tracker : IterationTracker
        **not set up**
    remove_init_outliers : float
        **not set up**
    """

    @property
    def minimum_frequency(self) -> float | dict[str, float]:
        if self.init_sampler is not None:
            return self.init_sampler.minimum_frequency
        else:
            raise AttributeError(
                "init_sampler not set. Cannot access minimum frequency."
            )

    @minimum_frequency.setter
    def minimum_frequency(self, value):
        if self.init_sampler is not None:
            self.init_sampler.minimum_frequency = value
        else:
            raise AttributeError(
                "init_sampler not set. Cannot update minimum frequency."
            )

    @property
    def maximum_frequency(self) -> float | dict[str, float]:
        if self.init_sampler is not None:
            return self.init_sampler.maximum_frequency
        else:
            raise AttributeError(
                "init_sampler not set. Cannot access maximum frequency."
            )

    @maximum_frequency.setter
    def maximum_frequency(self, value):
        if self.init_sampler is not None:
            self.init_sampler.maximum_frequency = value
        else:
            raise AttributeError(
                "init_sampler not set. Cannot update maximum frequency."
            )

    def _initialize_transforms(self):
        """
        Builds the transforms that are used in the GNPE loop.
        """
        data_settings = self.metadata["train_settings"]["data"]
        ifo_list = InterferometerList(data_settings["detectors"])

        gnpe_time_settings = data_settings.get("gnpe_time_shifts")
        gnpe_chirp_settings = data_settings.get("gnpe_chirp")
        gnpe_phase_settings = data_settings.get("gnpe_phase")
        if (
            not gnpe_time_settings
            and not gnpe_chirp_settings
            and not gnpe_phase_settings
        ):
            raise KeyError(
                "GNPE inference requires network trained for either chirp mass, "
                "coalescence time, or phase GNPE."
            )

        # transforms for gnpe loop, to be applied prior to sampling step:
        #   * reset the sample (e.g., clone non-gnpe transformed waveform)
        #   * blurring detector times to obtain gnpe proxies
        #   * shifting the strain by - gnpe proxies
        #   * repackaging & standardizing proxies to sample['context_parameters']
        #     for conditioning of the inference network
        transform_pre = []
        transform_pre.append(RenameKey("data", "waveform"))
        if gnpe_time_settings:
            transform_pre.append(
                GNPECoalescenceTimes(
                    ifo_list,
                    gnpe_time_settings["kernel"],
                    gnpe_time_settings["exact_equiv"],
                    inference=True,
                )
            )
            transform_pre.append(TimeShiftStrain(ifo_list, self.domain))
        transform_pre.append(
            SelectStandardizeRepackageParameters(
                {"context_parameters": data_settings["context_parameters"]},
                data_settings["standardization"],
                device=self.model.device,
            )
        )
        transform_pre.append(RenameKey("waveform", "data"))

        # Extract GNPE information (list of parameters, dict of kernels) from the
        # transforms.
        self.gnpe_parameters = []
        self.gnpe_kernel = PriorDict()
        for transform in transform_pre:
            if isinstance(transform, GNPEBase):
                self.gnpe_parameters += transform.input_parameter_names
                for k, v in transform.kernel.items():
                    self.gnpe_kernel[k] = v
        print("GNPE parameters: ", self.gnpe_parameters)
        print("GNPE kernel: ", self.gnpe_kernel)

        self.transform_pre = Compose(transform_pre)

        # transforms for gnpe loop, to be applied after sampling step:
        #   * de-standardization of parameters
        #   * post correction for geocent time (required for gnpe with exact equivariance)
        #   * computation of detectortimes from parameters (required for next gnpe
        #       iteration)
        self.transform_post = Compose(
            [
                SelectStandardizeRepackageParameters(
                    {"inference_parameters": self.inference_parameters},
                    data_settings["standardization"],
                    inverse=True,
                    as_type="dict",
                ),
                PostCorrectGeocentTime(),
                CopyToExtrinsicParameters(
                    "ra", "dec", "geocent_time", "chirp_mass", "mass_ratio", "phase"
                ),
                GetDetectorTimes(ifo_list, data_settings["ref_time"]),
            ]
        )

    def _kernel_log_prob(self, samples):
        # TODO: Reimplement as a method of GNPEBase.
        if len({"chirp_mass", "mass_ratio", "phase"} & self.gnpe_kernel.keys()) > 0:
            raise NotImplementedError("kernel log_prob only implemented for time gnpe.")
        gnpe_proxies_diff = {
            k: np.array(samples[k] - samples[f"{k}_proxy"])
            for k in self.gnpe_kernel.keys()
        }
        return self.gnpe_kernel.ln_prob(gnpe_proxies_diff, axis=0)


# Functions for frequency cropping. Used by Sampler classes and dingo-pipe.


def _validate_maximum_frequency(
    f_max: dict[str, float] | float,
    detectors: list[str],
    domain: UniformFrequencyDomain | MultibandedFrequencyDomain,
    crop_settings: dict | None,
):
    if isinstance(f_max, float):
        f_max = {d: f_max for d in detectors}
    if set(f_max) != set(detectors):
        raise ValueError(
            f"f_max must have exactly detectors {detectors}, got " f"{list(f_max)}."
        )
    f_max_vals = np.array([f_max[d] for d in detectors])

    # Hard upper bound
    if np.any(f_max_vals > domain.f_max):
        raise ValueError(f"f_max {f_max} > domain.f_max = {domain.f_max}.")

    # Nothing changed
    if np.all(f_max_vals == domain.f_max):
        return

    # Cropping must be on
    if not crop_settings or crop_settings.get("cropping_probability", 0.0) == 0.0:
        raise ValueError(
            f"Cropping disabled; cannot lower maximum frequency to {f_max}."
        )

    # Extract lower bounds
    floors = crop_settings.get("f_max_lower")
    if floors is None:
        floors = domain.f_max
    if not isinstance(floors, dict):
        floors = {d: floors for d in detectors}

    # Check lower bound.
    if not crop_settings.get("independent_detectors", True):
        if len(set(f_max_vals)) > 1:
            raise ValueError(
                f"Independent max frequencies per detector not enabled. "
                f"All frequencies must match, got f_max = {f_max}."
            )
        # TODO: Risk of non-constant floors with non-independent detectors.
        assert len(set(floors.values())) == 1
    for d in detectors:
        if f_max[d] < floors[d]:
            raise ValueError(
                f"Maximum frequency requested for {d} ({f_max[d]} Hz) "
                f"less than lower bound of {floors[d]} Hz."
            )


def _validate_minimum_frequency(
    f_min: dict[str, float] | float,
    detectors: list[str],
    domain: UniformFrequencyDomain | MultibandedFrequencyDomain,
    crop_settings: dict | None,
):
    if isinstance(f_min, float):
        f_min = {d: f_min for d in detectors}
    if set(f_min) != set(detectors):
        raise ValueError(
            f"f_min must have exactly detectors {detectors}, got {list(f_min)}."
        )
    f_min_vals = np.array([f_min[d] for d in detectors])

    # Hard lower bound
    if np.any(f_min_vals < domain.f_min):
        raise ValueError(f"f_min {f_min} < domain.f_min = {domain.f_min}.")

    # Nothing changed
    if np.all(f_min_vals == domain.f_min):
        return

    # Cropping must be on
    if not crop_settings or crop_settings.get("cropping_probability", 0.0) == 0.0:
        raise ValueError(
            f"Cropping disabled; cannot raise minimum frequency to {f_min}."
        )

    # Extract upper bounds
    caps = crop_settings.get("f_min_upper")
    if caps is None:
        caps = domain.f_min
    if not isinstance(caps, dict):
        caps = {d: caps for d in detectors}

    # Check upper bound.
    if not crop_settings.get("independent_detectors", True):
        if len(set(f_min_vals)) > 1:
            raise ValueError(
                f"Independent min frequencies per detector not enabled. "
                f"All frequencies must match, got f_min = {f_min}."
            )
        # TODO: Risk of non-constant caps with non-independent detectors.
        assert len(set(caps.values())) == 1
    for d in detectors:
        if f_min[d] > caps[d]:
            raise ValueError(
                f"Minimum frequency requested for {d} ({f_min[d]} Hz) "
                f"greater than upper bound of {caps[d]} Hz."
            )


def check_frequency_updates(
    model_metadata: dict,
    f_min: dict[str, float] | float | None = None,
    f_max: dict[str, float] | float | None = None,
):
    """
    Validate and apply optional minimum and maximum frequency constraints
    for a model’s frequency domain.

    This function checks that any provided per-detector minimum (`f_min`)
    or maximum (`f_max`) frequencies—either as a single float applied to
    all detectors or as a dict mapping each detector to its own value—:
      - Match exactly the set of detectors in the model metadata.
      - Respect the hard bounds defined by the domain (`domain.f_min` /
        `domain.f_max`).
      - Comply with optional random-strain-cropping settings (probability,
        independent vs. joint detectors, and per-detector caps/floors).

    Parameters
    ----------
    model_metadata : dict
        Dictionary containing the model’s training settings and data.
        Must include:
          - `["train_settings"]["data"]["detectors"]`: list of detector names.
          - `["train_settings"]["data"]["random_strain_cropping"]`: optional
            dict of cropping parameters.
    f_min : dict[str, float], float, or None, optional
        Single float or per-detector dict of minimum frequencies to enforce.
        If a float is provided, it is applied to all detectors. Each value
        must be ≥ `domain.f_min`. If `None`, no minimum-frequency
        validation is performed.
    f_max : dict[str, float], float, or None, optional
        Single float or per-detector dict of maximum frequencies to enforce.
        If a float is provided, it is applied to all detectors. Each value
        must be ≤ `domain.f_max`. If `None`, no maximum-frequency
        validation is performed.

    Raises
    ------
    ValueError
        - If `model_metadata` does not describe a `UniformFrequencyDomain`
          or `MultibandedFrequencyDomain`.
        - If `f_min`/`f_max` keys don’t exactly match the detector list.
        - If any requested frequency lies outside the hard domain bounds.
        - If cropping is disabled but a change in frequency is requested.
        - If per-detector constraints (independent vs. joint) or
          cropping caps/floors are violated.

    Returns
    -------
    None
    """
    crop_settings = model_metadata["train_settings"]["data"].get(
        "random_strain_cropping"
    )
    detectors = model_metadata["train_settings"]["data"]["detectors"]
    domain = build_domain_from_model_metadata(model_metadata, base=True)
    if not isinstance(domain, (UniformFrequencyDomain, MultibandedFrequencyDomain)):
        raise ValueError("Frequency updates only possible for frequency domains.")

    if f_min is not None:
        _validate_minimum_frequency(f_min, detectors, domain, crop_settings)
    if f_max is not None:
        _validate_maximum_frequency(f_max, detectors, domain, crop_settings)
